from .common import CaimanDataFrameExtensions, CaimanSeriesExtensions
from .mcorr import MCorrExtensions
from .cnmf import CNMFExtensions

__all__ = [
    "CaimanDataFrameExtensions",
    "CaimanSeriesExtensions",
    "MCorrExtensions",
    "CNMFExtensions",
]



import click
import numpy as np
import caiman as cm
from caiman.source_extraction.cnmf import cnmf as cnmf
from caiman.source_extraction.cnmf.params import CNMFParams
import psutil
import traceback
from pathlib import Path
from shutil import move as move_file
import os
import time
from datetime import datetime

if __name__ in ["__main__", "__mp_main__"]:  # when running in subprocess
    from mesmerize_core import set_parent_raw_data_path, load_batch
    from mesmerize_core.utils import IS_WINDOWS
else:  # when running with local backend
    from ..batch_utils import set_parent_raw_data_path, load_batch
    from ..utils import IS_WINDOWS


def run_algo(batch_path, uuid, data_path: str = None):
    algo_start = time.time()
    set_parent_raw_data_path(data_path)

    df = load_batch(batch_path)
    item = df[df["uuid"] == uuid].squeeze()

    input_movie_path = item["input_movie_path"]
    # resolve full path
    input_movie_path = str(df.paths.resolve(input_movie_path))

    output_dir = Path(batch_path).parent.joinpath(str(uuid))
    output_dir.mkdir(parents=True, exist_ok=True)

    params = item["params"]
    print("cnmfe params:", params)

    # adapted from current demo notebook
    if "MESMERIZE_N_PROCESSES" in os.environ.keys():
        try:
            n_processes = int(os.environ["MESMERIZE_N_PROCESSES"])
        except:
            n_processes = psutil.cpu_count() - 1
    else:
        n_processes = psutil.cpu_count() - 1
    # Start cluster for parallel processing
    c, dview, n_processes = cm.cluster.setup_cluster(
        backend="local", n_processes=n_processes, single_thread=False
    )

    try:
        fname_new = cm.save_memmap(
            [input_movie_path], base_name=f"{uuid}_cnmf-memmap_", order="C", dview=dview
        )

        print("making memmap")
        Yr, dims, T = cm.load_memmap(fname_new)
        images = np.reshape(Yr.T, [T] + list(dims), order="F")

        # TODO: if projections already exist from mcorr we don't
        #  need to waste compute time re-computing them here
        proj_paths = dict()
        for proj_type in ["mean", "std", "max"]:
            p_img = getattr(np, f"nan{proj_type}")(images, axis=0)
            proj_paths[proj_type] = output_dir.joinpath(
                f"{uuid}_{proj_type}_projection.npy"
            )
            np.save(str(proj_paths[proj_type]), p_img)

        d = dict()  # for output

        # force the CNMFE params
        cnmfe_params_dict = {
            "method_init": "corr_pnr",
            "n_processes": n_processes,
            "only_init": True,  # for 1p
            "center_psf": True,  # for 1p
            "normalize_init": False,  # for 1p
        }

        params_dict = {**cnmfe_params_dict, **params["main"]}

        cnmfe_params_dict = CNMFParams(params_dict=params_dict)
        cnm = cnmf.CNMF(
            n_processes=n_processes, dview=dview, params=cnmfe_params_dict
        )
        print("Performing CNMFE")
        cnm = cnm.fit(images)
        print("evaluating components")
        cnm.estimates.evaluate_components(images, cnm.params, dview=dview)

        cnmf_hdf5_path = output_dir.joinpath(f"{uuid}.hdf5").resolve()
        cnm.save(str(cnmf_hdf5_path))

        # save output paths to outputs dict
        d["cnmf-hdf5-path"] = cnmf_hdf5_path.relative_to(output_dir.parent)

        for proj_type in proj_paths.keys():
            d[f"{proj_type}-projection-path"] = proj_paths[proj_type].relative_to(
                output_dir.parent
            )

        cnmf_memmap_path = output_dir.joinpath(Path(fname_new).name)
        if IS_WINDOWS:
            Yr._mmap.close()  # accessing private attr but windows is annoying otherwise
        move_file(fname_new, cnmf_memmap_path)

        cnmfe_memmap_path = cnmf_memmap_path.relative_to(output_dir.parent)

        d.update(
            {
                "cnmf-memmap-path": cnmfe_memmap_path,
                "success": True,
                "traceback": None,
            }
        )

    except:
        d = {"success": False, "traceback": traceback.format_exc()}

    cm.stop_server(dview=dview)

    # Add dictionary to output column of series
    df.loc[df["uuid"] == uuid, "outputs"] = [d]
    # Add ran timestamp to ran_time column of series
    df.loc[df["uuid"] == uuid, "ran_time"] = datetime.now().isoformat(timespec="seconds", sep="T")
    df.loc[df["uuid"] == uuid, "algo_duration"] = str(round(time.time() - algo_start, 2)) + " sec"
    # save dataframe to disc
    df.to_pickle(batch_path)


@click.command()
@click.option("--batch-path", type=str)
@click.option("--uuid", type=str)
@click.option("--data-path")
def main(batch_path, uuid, data_path: str = None):
    run_algo(batch_path, uuid, data_path)


if __name__ == "__main__":
    main()



class BatchItemNotRunError(Exception):
    pass


class BatchItemUnsuccessfulError(Exception):
    pass


class WrongAlgorithmExtensionError(Exception):
    pass


class DependencyError(Exception):
    pass



import os
from pathlib import Path
from typing import Union

import pandas as pd

from .utils import validate_path

CURRENT_BATCH_PATH: Path = None  # only one batch at a time
PARENT_DATA_PATH: Path = None

COMPUTE_BACKEND_SUBPROCESS = "subprocess"  #: subprocess backend
COMPUTE_BACKEND_SLURM = "slurm"  #: SLURM backend, not yet implemented
COMPUTE_BACKEND_LOCAL = "local"

COMPUTE_BACKENDS = [COMPUTE_BACKEND_SUBPROCESS, COMPUTE_BACKEND_SLURM, COMPUTE_BACKEND_LOCAL]

DATAFRAME_COLUMNS = ["algo", "item_name", "input_movie_path", "params", "outputs", "added_time", "ran_time", "algo_duration", "comments", "uuid"]


def set_parent_raw_data_path(path: Union[Path, str]) -> Path:
    """
    Set the global `PARENT_DATA_PATH`

    Parameters
    ----------
    path: Path or str
        Full parent data path
    """
    global PARENT_DATA_PATH
    path = Path(validate_path(path))
    if not path.is_dir():
        raise NotADirectoryError(
            "The directory passed to `set_parent_raw_data_path()` does not exist.\n"
        )
    PARENT_DATA_PATH = path

    return PARENT_DATA_PATH


def get_parent_raw_data_path() -> Path:
    """
    Get the global `PARENT_DATA_PATH`

    Returns
    -------
    Path
        global `PARENT_DATA_PATH` as a Path object

    """
    global PARENT_DATA_PATH
    return PARENT_DATA_PATH


class _BasePathExtensions:
    def __init__(self, data: Union[pd.DataFrame, pd.Series]):
        self._data = data

    def set_batch_path(self, path: Union[str, Path]):
        self._data.attrs["batch_path"] = Path(path)

    def get_batch_path(self) -> Path:
        """
        Get the full path to the batch dataframe file

        Returns
        -------
        Path
            full path to the batch dataframe file as a Path object
        """
        if "batch_path" in self._data.attrs.keys():
            if self._data.attrs["batch_path"] is not None:
                return self._data.attrs["batch_path"]
        else:
            raise ValueError("Batch path is not set")

    def resolve(self, path: Union[str, Path]) -> Path:
        """
        Resolve the full path of the passed ``path`` if possible, first tries
        "batch_dir" then "raw_data_dir".

        Parameters
        ----------
        path: str or Path
            The relative path to resolve

        Returns
        -------
        Path
            Full path with the batch path or raw data path appended

        """
        path = Path(path)
        # check if input movie is within batch dir
        if self.get_batch_path().parent.joinpath(path).exists():
            return self.get_batch_path().parent.joinpath(path)

        # else check if in parent raw data dir
        elif get_parent_raw_data_path() is not None:
            if get_parent_raw_data_path().joinpath(path).exists():
                return get_parent_raw_data_path().joinpath(path)

        raise FileNotFoundError(f"Could not resolve full path of:\n{path}")

    def split(self, path: Union[str, Path]):
        """
        Split a full path into (batch_dir, relative_path) or (raw_data_dir, relative_path)

        Parameters
        ----------
        path: str or Path
            Full path to split with respect to batch_dir or raw_data_dir

        Returns
        -------
        Tuple[Path, Path]
            (<batch_dir> or <raw_data_dir>, <relative_path>)

        """
        path = Path(path)
        # check if input movie is within batch dir
        if self.get_batch_path().parent in path.parents:
            return self.get_batch_path().parent, path.relative_to(
                self.get_batch_path().parent
            )

        # else check if in parent raw data dir
        elif get_parent_raw_data_path() is not None:
            if get_parent_raw_data_path() in path.parents:
                return get_parent_raw_data_path(), path.relative_to(
                    get_parent_raw_data_path()
                )

        raise NotADirectoryError(
                f"Could not split `path`:\n{path}"
                f"\nnot relative to either batch path:\n{self.get_batch_path()}"
                f"\nor parent raw data path:\n{get_parent_raw_data_path()}"
            )


@pd.api.extensions.register_dataframe_accessor("paths")
class PathsDataFrameExtension(_BasePathExtensions):
    pass


@pd.api.extensions.register_series_accessor("paths")
class PathsSeriesExtension(_BasePathExtensions):
    pass


def load_batch(path: Union[str, Path]) -> pd.DataFrame:
    """
    Load the batch dataframe pickle file

    Parameters
    ----------
    path: str or Path

    Returns
    -------
    pd.DataFrame
        batch dataframe loaded from the specified path

    Examples
    --------

    .. code-block:: python

        from mesmerize_core import *

        df = load_batch("/path/to/batch.pickle")

        # view dataframe
        df.head()

    """

    path = validate_path(path)

    df = pd.read_pickle(Path(path))

    df.paths.set_batch_path(path)

    # check to see if added and ran timestamp are in df
    if all(item in df.columns for item in ["added_time", "ran_time", "algo_duration"]):
        return df
    else:
        df["added_time"] = None
        df["ran_time"] = None
        df["algo_duration"] = None
        return df


def create_batch(path: Union[str, Path], remove_existing: bool = False) -> pd.DataFrame:
    """
    Create a new batch DataFrame

    Parameters
    ----------
    path: str or Path
        path to save the new batch DataFrame as a pickle file

    remove_existing: bool
        If ``True``, remove an existing batch DataFrame file if it exists at the given `path`, default ``False``

    Returns
    -------
    pd.DataFrame
        New empty batch DataFrame

    Examples
    --------

    .. code-block:: python

        from mesmerize_core import *

        df = create_batch("/path/to/new_batch.pickle")

    """
    path = validate_path(path)

    if Path(path).is_file():
        if remove_existing:
            os.remove(path)
        else:
            raise FileExistsError(
                f"Batch file already exists at specified location: {path}"
            )

    if not Path(path).parent.is_dir():
        os.makedirs(Path(path).parent)

    df = pd.DataFrame(columns=DATAFRAME_COLUMNS)
    df.paths.set_batch_path(path)

    df.to_pickle(path)

    return df


def get_full_raw_data_path(path: Union[Path, str]) -> Path:
    path = Path(path)
    if PARENT_DATA_PATH is not None:
        return PARENT_DATA_PATH.joinpath(path)

    return path



import traceback
import click
import caiman as cm
from caiman.source_extraction.cnmf.params import CNMFParams
from caiman.motion_correction import MotionCorrect
from caiman.summary_images import local_correlations_movie_offline
import psutil
import pandas as pd
import os
from pathlib import Path
import numpy as np
from shutil import move as move_file
import time
from datetime import datetime


# prevent circular import
if __name__ in ["__main__", "__mp_main__"]:  # when running in subprocess
    from mesmerize_core import set_parent_raw_data_path, load_batch
else:  # when running with local backend
    from ..batch_utils import set_parent_raw_data_path, load_batch


def run_algo(batch_path, uuid, data_path: str = None):
    algo_start = time.time()
    set_parent_raw_data_path(data_path)

    batch_path = Path(batch_path)
    df = load_batch(batch_path)

    item = df[df["uuid"] == uuid].squeeze()
    # resolve full path
    input_movie_path = str(df.paths.resolve(item["input_movie_path"]))

    # because caiman doesn't let you specify filename to save memmap files
    # create dir with uuid as the dir item_name
    output_dir = Path(batch_path).parent.joinpath(str(uuid))
    caiman_temp_dir = str(output_dir)
    os.makedirs(caiman_temp_dir, exist_ok=True)
    os.environ["CAIMAN_TEMP"] = caiman_temp_dir
    os.environ["CAIMAN_NEW_TEMPFILE"] = "True"

    params = item["params"]

    # adapted from current demo notebook
    if "MESMERIZE_N_PROCESSES" in os.environ.keys():
        try:
            n_processes = int(os.environ["MESMERIZE_N_PROCESSES"])
        except:
            n_processes = psutil.cpu_count() - 1
    else:
        n_processes = psutil.cpu_count() - 1

    print("starting mc")
    # Start cluster for parallel processing
    c, dview, n_processes = cm.cluster.setup_cluster(
        backend="local", n_processes=n_processes, single_thread=False
    )

    rel_params = dict(params["main"])
    opts = CNMFParams(params_dict=rel_params)
    # Run MC, denote boolean 'success' if MC completes w/out error
    try:
        # Run MC
        fnames = [input_movie_path]
        mc = MotionCorrect(fnames, dview=dview, **opts.get_group("motion"))
        mc.motion_correct(save_movie=True)

        # find path to mmap file
        memmap_output_path_temp = df.paths.resolve(mc.mmap_file[0])

        # filename to move the output back to data dir
        mcorr_memmap_path = output_dir.joinpath(
            f"{uuid}-{memmap_output_path_temp.name}"
        )

        # move the output file
        move_file(memmap_output_path_temp, mcorr_memmap_path)

        print("mc finished successfully!")

        print("computing projections")
        Yr, dims, T = cm.load_memmap(str(mcorr_memmap_path))
        images = np.reshape(Yr.T, [T] + list(dims), order="F")

        proj_paths = dict()
        for proj_type in ["mean", "std", "max"]:
            p_img = getattr(np, f"nan{proj_type}")(images, axis=0)
            proj_paths[proj_type] = output_dir.joinpath(
                f"{uuid}_{proj_type}_projection.npy"
            )
            np.save(str(proj_paths[proj_type]), p_img)

        print("Computing correlation image")
        Cns = local_correlations_movie_offline(
            [str(mcorr_memmap_path)],
            remove_baseline=True,
            window=1000,
            stride=1000,
            winSize_baseline=100,
            quantil_min_baseline=10,
            dview=dview,
        )
        Cn = Cns.max(axis=0)
        Cn[np.isnan(Cn)] = 0
        cn_path = output_dir.joinpath(f"{uuid}_cn.npy")
        np.save(str(cn_path), Cn, allow_pickle=False)

        # output dict for pandas series for dataframe row
        d = dict()

        print("finished computing correlation image")

        # Compute shifts
        if params["main"]["pw_rigid"] == True:
            x_shifts = mc.x_shifts_els
            y_shifts = mc.y_shifts_els
            shifts = [x_shifts, y_shifts]
            shift_path = output_dir.joinpath(f"{uuid}_shifts.npy")
            np.save(str(shift_path), shifts)
        else:
            shifts = mc.shifts_rig
            shift_path = output_dir.joinpath(f"{uuid}_shifts.npy")
            np.save(str(shift_path), shifts)

        # relative paths
        cn_path = cn_path.relative_to(output_dir.parent)
        mcorr_memmap_path = mcorr_memmap_path.relative_to(output_dir.parent)
        shift_path = shift_path.relative_to(output_dir.parent)
        for proj_type in proj_paths.keys():
            d[f"{proj_type}-projection-path"] = proj_paths[proj_type].relative_to(
                output_dir.parent
            )

        d.update(
            {
                "mcorr-output-path": mcorr_memmap_path,
                "corr-img-path": cn_path,
                "shifts": shift_path,
                "success": True,
                "traceback": None,
            }
        )

    except:
        d = {"success": False, "traceback": traceback.format_exc()}
        print("mc failed, stored traceback in output")

    cm.stop_server(dview=dview)

    # Add dictionary to output column of series
    df.loc[df["uuid"] == uuid, "outputs"] = [d]
    # Add ran timestamp to ran_time column of series
    df.loc[df["uuid"] == uuid, "ran_time"] = datetime.now().isoformat(timespec="seconds", sep="T")
    df.loc[df["uuid"] == uuid, "algo_duration"] = str(round(time.time() - algo_start, 2)) + " sec"
    # Save DataFrame to disk
    df.to_pickle(batch_path)


@click.command()
@click.option("--batch-path", type=str)
@click.option("--uuid", type=str)
@click.option("--data-path", type=str)
def main(batch_path, uuid, data_path: str = None):
    run_algo(batch_path, uuid, data_path)


if __name__ == "__main__":
    main()



from warnings import warn
from abc import ABC, abstractmethod
from pathlib import Path
from typing import *

import numpy as np

slice_or_int_or_range = Union[int, slice, range]


class LazyArray(ABC):
    """
    Base class for arrays that exhibit lazy computation upon indexing
    """
    @property
    @abstractmethod
    def dtype(self) -> str:
        """
        str
            data type
        """
        pass

    @property
    @abstractmethod
    def shape(self) -> Tuple[int, int, int]:
        """
        Tuple[int]
            (n_frames, dims_x, dims_y)
        """
        pass

    @property
    @abstractmethod
    def n_frames(self) -> int:
        """
        int
            number of frames
        """
        pass

    @property
    @abstractmethod
    def min(self) -> float:
        """
        float
            min value of the array if it were fully computed
        """
        pass

    @property
    @abstractmethod
    def max(self) -> float:
        """
        float
            max value of the array if it were fully computed
        """
        pass

    @property
    def ndim(self) -> int:
        """
        int
            Number of dimensions
        """
        return len(self.shape)

    @property
    def nbytes(self) -> int:
        """
        int
            number of bytes for the array if it were fully computed
        """
        return np.prod(self.shape + (np.dtype(self.dtype).itemsize,))

    @property
    def nbytes_gb(self) -> float:
        """
        float
            number of gigabytes for the array if it were fully computed
        """
        return self.nbytes / 1e9

    @abstractmethod
    def _compute_at_indices(self, indices: Union[int, slice]) -> np.ndarray:
        """
        Lazy computation logic goes here. Computes the array at the desired indices.

        Parameters
        ----------
        indices: Union[int, slice]
            the user's desired slice, i.e. slice object or int passed from `__getitem__()`

        Returns
        -------
        np.ndarray
            array at the indexed slice
        """
        pass

    def as_numpy(self):
        """
        NOT RECOMMENDED, THIS COULD BE EXTREMELY LARGE. Converts to a standard numpy array in RAM.

        Returns
        -------
        np.ndarray
        """
        warn(
            f"\nYou are trying to create a numpy.ndarray from a LazyArray, "
            f"this is not recommended and could take a while.\n\n"
            f"Estimated size of final numpy array: "
            f"{self.nbytes_gb:.2f} GB"
        )
        a = np.zeros(shape=self.shape, dtype=self.dtype)

        for i in range(self.n_frames):
            a[i] = self[i]

        return a

    def save_hdf5(self, filename: Union[str, Path]):
        pass

    def __getitem__(
            self,
            item: Union[int, Tuple[slice_or_int_or_range]]
    ):
        if isinstance(item, int):
            indexer = item

        # numpy int scaler
        elif isinstance(item, np.integer):
            indexer = item.item()

        # treat slice and range the same
        elif isinstance(item, (slice, range)):
            indexer = item

        elif isinstance(item, tuple):
            if len(item) > len(self.shape):
                raise IndexError(
                    f"Cannot index more dimensions than exist in the array. "
                    f"You have tried to index with <{len(item)}> dimensions, "
                    f"only <{len(self.shape)}> dimensions exist in the array"
                )

            indexer = item[0]

        else:
            raise IndexError(
                f"You can index LazyArrays only using slice, int, or tuple of slice and int, "
                f"you have passed a: <{type(item)}>"
            )

        # treat slice and range the same
        if isinstance(indexer, (slice, range)):
            start = indexer.start
            stop = indexer.stop
            step = indexer.step

            if start is not None:
                if start > self.n_frames:
                    raise IndexError(f"Cannot index beyond `n_frames`.\n"
                                     f"Desired frame start index of <{start}> "
                                     f"lies beyond `n_frames` <{self.n_frames}>")
            if stop is not None:
                if stop > self.n_frames:
                    raise IndexError(f"Cannot index beyond `n_frames`.\n"
                                     f"Desired frame stop index of <{stop}> "
                                     f"lies beyond `n_frames` <{self.n_frames}>")

            if step is None:
                step = 1
            
            # convert indexer to slice if it was a range, allows things like decord.VideoReader slicing
            indexer = slice(start, stop, step)  # in case it was a range object
            
            # dimension_0 is always time
            frames = self._compute_at_indices(indexer)

            # index the remaining dims after lazy computing the frame(s)
            if isinstance(item, tuple):
                if len(item) == 2:
                    return frames[:, item[1]]
                elif len(item) == 3:
                    return frames[:, item[1], item[2]]

            else:
                return frames

        elif isinstance(indexer, int):
            return self._compute_at_indices(indexer)

    def __repr__(self):
        return f"{self.__class__.__name__} @{hex(id(self))}\n" \
               f"{self.__class__.__doc__}\n" \
               f"Frames are computed only upon indexing\n" \
               f"shape [frames, x, y]: {self.shape}\n"



from typing import *
from pathlib import Path
from warnings import warn

import numpy as np

try:
    from decord import VideoReader
except ImportError:
    HAS_DECORD = False
else:
    HAS_DECORD = True

from ._base import LazyArray


class LazyVideo(LazyArray):
    def __init__(
        self, path: Union[Path, str],
            min_max: Tuple[int, int] = None,
            as_grayscale: bool = False,
            rgb_weights: Tuple[float, float, float] = (0.299, 0.587, 0.114),
            **kwargs,
    ):
        """
        LazyVideo reader, basically just a wrapper for ``decord.VideoReader``.
        Should support opening anything that decord can open.

        **Important:** requires ``decord`` to be installed: https://github.com/dmlc/decord

        Parameters
        ----------
        path: Path or str
            path to video file

        min_max: Tuple[int, int], optional
            min and max vals of the entire video, uses min and max of 10th frame if not provided

        as_grayscale: bool, optional
            return grayscale frames upon slicing

        rgb_weights: Tuple[float, float, float], optional
            (r, g, b) weights used for grayscale conversion if ``as_graycale`` is ``True``.
            default is (0.299, 0.587, 0.114)

        kwargs
            passed to ``decord.VideoReader``

        Examples
        --------

        Lazy loading with CPU

        .. code-block:: python

            from mesmerize_core.arrays import LazyVideo

            vid = LazyVideo("path/to/video.mp4")

            # use fpl to visualize

            import fastplotlib as fpl

            iw = fpl.ImageWidget(vid)
            iw.show()


        Lazy loading with GPU, decord must be compiled with CUDA options to use this

        .. code-block:: python

            from decord import gpu
            from mesmerize_core.arrays import LazyVideo

            gpu_context = gpu(0)

            vid = LazyVideo("path/to/video.mp4", ctx=gpu_context)

        """
        if not HAS_DECORD:
            raise ImportError("You must install `decord` to use LazyVideo")

        self._video_reader = VideoReader(str(path), **kwargs)

        try:
            frame0 = self._video_reader[10].asnumpy()
        except IndexError:
            frame0 = self._video_reader[0].asnumpy()

        self._shape = (self._video_reader._num_frame, *frame0.shape[:-1])

        if len(frame0.shape) > 2:
            # we assume the shape of a frame is [x, y, RGB]
            self._is_color = True
        else:
            # we assume is already grayscale
            self._is_color = False

        self._dtype = frame0.dtype

        if min_max is not None:
            self._min, self._max = min_max
        else:
            self._min = frame0.min()
            self._max = frame0.max()

        self.as_grayscale = as_grayscale
        self.rgb_weights = rgb_weights

    @property
    def dtype(self) -> str:
        return self._dtype

    @property
    def shape(self) -> Tuple[int, int, int]:
        """[n_frames, x, y], RGB color dim not included in shape"""
        return self._shape

    @property
    def n_frames(self) -> int:
        return self.shape[0]

    @property
    def min(self) -> float:
        warn("min not implemented for LazyTiff, returning min of 0th index")
        return self._min

    @property
    def max(self) -> float:
        warn("max not implemented for LazyTiff, returning min of 0th index")
        return self._max

    def _compute_at_indices(self, indices: Union[int, slice]) -> np.ndarray:
        if not self.as_grayscale:
            return self._video_reader[indices].asnumpy()

        if self._is_color:
            a = self._video_reader[indices].asnumpy()

            # R + G + B -> grayscale
            gray = a[..., 0] * self.rgb_weights[0] +\
            a[..., 1] * self.rgb_weights[1] +\
            a[..., 2] * self.rgb_weights[2]

            return gray

        warn("Video is already grayscale, just returning")
        return self._video_reader[indices].asnumpy()



from ._cnmf import LazyArrayRCM, LazyArrayRCB, LazyArrayResiduals
from ._tiff import LazyTiff
from ._video import LazyVideo

__all__ = [
    "LazyArrayRCM",
    "LazyArrayRCB",
    "LazyArrayResiduals",
    "LazyTiff",
    "LazyVideo"
]



import tifffile
from pathlib import Path
import numpy as np
from caiman import load_memmap

from .utils import warning_experimental
from .arrays import LazyTiff

try:
    import pims
    HAS_PIMS = True
except (ModuleNotFoundError, ImportError):
    HAS_PIMS = False


def default_reader(path: str, **kwargs):
    ext = Path(path).suffixes[-1]
    if ext in [".tiff", ".tif", ".btf"]:
        try:
            movie = tiff_memmap_reader(path, **kwargs)
        except:  # if file is not memmapable
            movie = tiff_lazyarray(path, **kwargs)

        return movie

    if ext in [".mmap", ".memmap"]:
        return caiman_memmap_reader(path, **kwargs)

    else:
        raise ValueError(
            f"No default movie reader for given file extension: '{ext}'"
        )


def tiff_memmap_reader(path: str, **kwargs) -> np.memmap:
    return tifffile.memmap(path, **kwargs)


@warning_experimental("This feature is new and might change in the future")
def tiff_lazyarray(path: str) -> LazyTiff:
    # random access speed on a magnetic HDD is ~30Hz for simultaneously slicing of 20 frames from Teena's tiff files
    # much slower than tifffile.memmap but this is just a last resort anyways
    return LazyTiff(path)


def caiman_memmap_reader(path: str, **kwargs) -> np.memmap:
    Yr, dims, T = load_memmap(path, **kwargs)
    return np.reshape(Yr.T, [T] + list(dims), order="F")


def pims_reader(path: str, **kwargs):
    if not HAS_PIMS:
        raise ModuleNotFoundError(
            "you must install `pims` to use the pims reader"
        )
    return pims.open(path, **kwargs)



from itertools import product as iter_product
from typing import *
from time import time
from warnings import warn

import numpy as np
from scipy.sparse import csc_matrix

from ._base import LazyArray


class LazyArrayRCM(LazyArray):
    """LazyArray for reconstructed movie, i.e. A ⊗ C"""
    def __init__(
            self,
            spatial: np.ndarray,
            temporal: np.ndarray,
            frame_dims: Tuple[int, int],
    ):
        """
        Parameters
        ----------
        spatial: np.ndarray
            spatial components

        temporal: np.ndarray
            temporal components

        frame_dims: Tuple[int, int]
            frame dimensions
            
        """

        if spatial.shape[1] != temporal.shape[0]:
            raise ValueError(
                f"Number of temporal components provided: `{temporal.shape[0]}` "
                f"does not equal number of spatial components provided: `{spatial.shape[1]}`"
            )

        self._spatial = spatial
        self._temporal = temporal

        self._shape: Tuple[int, int, int] = (temporal.shape[1], *frame_dims)

        # determine dtype
        if self.spatial.dtype == self.temporal.dtype:
            self._dtype = self.temporal.dtype.name
        else:
            self._dtype = self[0].dtype.name

        # precompute min and max vals for each component for spatial and temporal
        temporal_max = np.nanmax(self.temporal, axis=1)
        temporal_min = np.nanmin(self.temporal, axis=1)

        if isinstance(self.spatial, csc_matrix):
            spatial_max = self.spatial.max(axis=0).toarray()
            spatial_min = self.spatial.min(axis=0).toarray()
        else:
            spatial_max = self.spatial.max(axis=0)
            spatial_min = self.spatial.min(axis=0)

        prods = list()
        for t, s in iter_product([temporal_min, temporal_max], [spatial_min, spatial_max]):
            _p = np.multiply(t, s)
            prods.append(np.nanmin(_p))
            prods.append(np.nanmax(_p))
        
        self._max = np.max(prods)
        self._min = np.min(prods)

        temporal_mean = np.nanmean(self.temporal, axis=1)
        temporal_std = np.nanstd(self.temporal, axis=1)

        self._mean_image = self.spatial.dot(temporal_mean).reshape(frame_dims, order="F")
        self._max_image = self.spatial.dot(temporal_max).reshape(frame_dims, order="F")
        self._min_image = self.spatial.dot(temporal_min).reshape(frame_dims, order="F")
        self._std_image = self.spatial.dot(temporal_std).reshape(frame_dims, order="F")

    @property
    def spatial(self) -> np.ndarray:
        return self._spatial

    @property
    def temporal(self) -> np.ndarray:
        return self._temporal

    @property
    def n_components(self) -> int:
        return self._spatial.shape[1]

    @property
    def n_frames(self) -> int:
        return self._temporal.shape[1]

    @property
    def shape(self) -> Tuple[int, int, int]:
        return self._shape

    @property
    def dtype(self) -> str:
        return self._dtype

    @property
    def min(self) -> float:
        return self._min

    @property
    def max(self) -> float:
        return self._max

    @property
    def mean_image(self) -> np.ndarray:
        """mean projection image"""
        return self._mean_image

    @property
    def max_image(self) -> np.ndarray:
        """max projection image"""
        return self._max_image

    @property
    def min_image(self) -> np.ndarray:
        """min projection image"""
        return self._min_image

    @property
    def std_image(self) -> np.ndarray:
        """standard deviation projection image"""
        return self._std_image
    
    def _compute_at_indices(self, indices: Union[int, Tuple[int, int]]) -> np.ndarray:
        rcm = self.spatial.dot(
            self.temporal[:, indices]
        ).reshape(
            self.shape[1:] + (-1,), order="F"
        ).transpose([2, 0, 1])

        if rcm.shape[0] == 1:
            return rcm[0]  # 2d single frame
        else:
            return rcm

    def __repr__(self):
        r = super().__repr__()
        return f"{r}" \
               f"n_components: {self.n_components}"

    def __eq__(self, other):
        if not isinstance(other, LazyArrayRCM):
            raise TypeError(f"cannot compute equality for against types that are not {self.__class__.__name__}")

        if (self.spatial == other.spatial) and (self.temporal == other.temporal):
            return True
        else:
            return False


# implementation for reconstructed background is identical
# this is just an interface to separate them
class LazyArrayRCB(LazyArrayRCM):
    """Lazy array for reconstructed background, i.e. b ⊗ f"""


class LazyArrayResiduals(LazyArray):
    """Lazy array for residuals, i.e. Y - (A ⊗ C) - (b ⊗ f)"""
    def __init__(
            self,
            raw_movie: np.ndarray,
            rcm: LazyArrayRCM,
            rcb: LazyArrayRCB,
            timeout: int = 10
    ):
        """
        Create a LazyArray of the residuals, ``Y - (A ⊗ C) - (b ⊗ f)``

        Parameters
        ----------
        raw_movie: np.memmap
            numpy memmap of the raw movie

        rcm: LazyArrayRCM
            reconstructed movie lazy array

        rcb: LazyArrayRCB
            reconstructed background lazy array

        timeout: int, default ``10``
            timeout for min-max calculation, not implemented yet

        """
        self._raw_movie = raw_movie
        self._rcm = rcm
        self._rcb = rcb

        # I was lazy to figure out how to copy tuples
        self._shape = (
            self._raw_movie.shape[0],
            self._raw_movie.shape[1],
            self._raw_movie.shape[2],
        )

        if self._raw_movie.dtype == self._rcm.dtype == self._rcb.dtype:
            self._dtype = self._raw_movie.dtype
        else:
            self._dtype = self[0].dtype.name

        # TODO: implement min max for residuals
        # min_max_raw = self._quick_min_max(raw_movie, timeout)
        # if min_max_raw is None:
        #     self._min = None
        #     self._max = None

        # else:
        #     _min, _max = min_max_raw
        #
        #     _min = _min - self._rcm.max - self._rcb.max
        #     _max = _max -

    def _quick_min_max(self, data, timeout):
        # adapted from pyqtgraph.ImageView
        # Estimate the min/max values of *data* by subsampling.
        # Returns [(min, max), ...] with one item per channel

        t = time()
        while data.size > 1e6:
            ax = np.argmax(data.shape)
            sl = [slice(None)] * data.ndim
            sl[ax] = slice(None, None, 2)
            data = data[tuple(sl)]
            if (time() - t) > timeout:
                return None

        return float(np.nanmin(data)), float(np.nanmax(data))

    @property
    def dtype(self) -> str:
        return self._dtype

    @property
    def shape(self) -> Tuple[int, int, int]:
        return self._shape

    @property
    def n_frames(self) -> int:
        return self._shape[0]

    # TODO: implement min max for residuals
    @property
    def min(self) -> float:
        warn("min and max not yet implemented for LazyArrayResiduals. "
             "Using first frame of raw movie")
        return float(self._raw_movie[0].min())

    @property
    def max(self) -> float:
        warn("min and max not yet implemented for LazyArrayResiduals. "
             "Using first frame of raw movie")
        return float(self._raw_movie[0].max())

    def _compute_at_indices(self, indices: Union[int, slice]) -> np.ndarray:
        residuals = self._raw_movie[indices] - self._rcm[indices] - self._rcb[indices]
        return residuals



from functools import wraps
from typing import Union, Optional

import pandas as pd
import time
import numpy as np
import sys
from caiman.source_extraction.cnmf import CNMF
import re
from sys import getsizeof
import copy


def _check_arg_equality(args, cache_args):
    if not type(args) == type(cache_args):
        return False
    if isinstance(cache_args, np.ndarray):
        return np.array_equal(cache_args, args)
    else:
        return cache_args == args


def _check_args_equality(args, cache_args):
    if len(args) != len(cache_args):
        return False
    equality = list()
    if isinstance(args, tuple):
        for arg, cache_arg in zip(args, cache_args):
            equality.append(_check_arg_equality(arg, cache_arg))
    else:
        for k in args.keys():
            equality.append(_check_arg_equality(args[k], cache_args[k]))
    return all(equality)


def _return_wrapper(output, copy_bool):
    if copy_bool == True:
        return copy.deepcopy(output)
    else:
        return output


class Cache:
    def __init__(self, cache_size: Optional[Union[int, str]] = None):
        self.cache = pd.DataFrame(
            data=None,
            columns=["uuid", "function", "args", "kwargs", "return_val", "time_stamp"],
        )
        self.set_maxsize(cache_size)

    def get_cache(self):
        return self.cache

    def clear_cache(self):
        while len(self.cache.index) != 0:
            self.cache.drop(index=self.cache.index[-1], axis=0, inplace=True)

    def set_maxsize(self, max_size: Union[int, str]):
        if max_size is None:
            self.storage_type = "RAM"
            self.size = 1024**3
        elif isinstance(max_size, str):
            self.storage_type = "RAM"
            if max_size.endswith("G"):
                self.size = int(max_size[:-1]) * 1024**3
            elif max_size.endswith("M"):
                self.size = int(max_size[:-1]) * 1024**2
        else:
            self.storage_type = "ITEMS"
            self.size = max_size

    def _get_cache_size_bytes(self):
        """Returns in bytes"""
        cache_size = 0
        for i in range(len(self.cache.index)):
            if isinstance(self.cache.iloc[i, 4], np.ndarray):
                cache_size += self.cache.iloc[i, 4].data.nbytes
            elif isinstance(self.cache.iloc[i, 4], (tuple, list)):
                for lists in self.cache.iloc[i, 4]:
                    for array in lists:
                        cache_size += array.data.nbytes
            elif isinstance(self.cache.iloc[i, 4], CNMF):
                sizes = list()
                for attr in self.cache.iloc[i, 4].estimates.__dict__.values():
                    if isinstance(attr, np.ndarray):
                        sizes.append(attr.data.nbytes)
                    else:
                        sizes.append(getsizeof(attr))
            else:
                cache_size += sys.getsizeof(self.cache.iloc[i, 4])

        return cache_size

    def use_cache(self, func):
        @wraps(func)
        def _use_cache(instance, *args, **kwargs):
            if "return_copy" in kwargs.keys():
                return_copy = kwargs["return_copy"]
            else:
                return_copy = True

            if self.size == 0:
                self.clear_cache()
                return _return_wrapper(func(instance, *args, **kwargs), return_copy)

            # if cache is empty, will always be a cache miss
            if len(self.cache.index) == 0:
                return_val = func(instance, *args, **kwargs)
                self.cache.loc[len(self.cache.index)] = [
                    instance._series["uuid"],
                    func.__name__,
                    args,
                    kwargs,
                    return_val,
                    time.time(),
                ]
                return _return_wrapper(return_val, copy_bool=return_copy)

            # checking to see if there is a cache hit
            for i in range(len(self.cache.index)):
                if (
                    self.cache.iloc[i, 0] == instance._series["uuid"]
                    and self.cache.iloc[i, 1] == func.__name__
                    and _check_args_equality(args, self.cache.iloc[i, 2])
                    and _check_args_equality(kwargs, self.cache.iloc[i, 3])
                ):
                    self.cache.iloc[i, 5] = time.time()
                    return_val = self.cache.iloc[i, 4]
                    return _return_wrapper(self.cache.iloc[i, 4], copy_bool=return_copy)

            # no cache hit, must check cache limit, and if limit is going to be exceeded...remove least recently used and add new entry
            # if memory type is 'ITEMS': drop the least recently used and then add new item
            if self.storage_type == "ITEMS" and len(self.cache.index) >= self.size:
                return_val = func(instance, *args, **kwargs)
                self.cache.drop(
                    index=self.cache.sort_values(
                        by=["time_stamp"], ascending=False
                    ).index[-1],
                    axis=0,
                    inplace=True,
                )
                self.cache = self.cache.reset_index(drop=True)
                self.cache.loc[len(self.cache.index)] = [
                    instance._series["uuid"],
                    func.__name__,
                    args,
                    kwargs,
                    return_val,
                    time.time(),
                ]
                return _return_wrapper(
                    self.cache.iloc[len(self.cache.index) - 1, 4], copy_bool=return_copy
                )
            # if memory type is 'RAM': add new item and then remove least recently used items until cache is under correct size again
            elif self.storage_type == "RAM":
                while self._get_cache_size_bytes() > self.size:
                    self.cache.drop(
                        index=self.cache.sort_values(
                            by=["time_stamp"], ascending=False
                        ).index[-1],
                        axis=0,
                        inplace=True,
                    )
                    self.cache = self.cache.reset_index(drop=True)
                return_val = func(instance, *args, **kwargs)
                self.cache.loc[len(self.cache.index)] = [
                    instance._series["uuid"],
                    func.__name__,
                    args,
                    kwargs,
                    return_val,
                    time.time(),
                ]
            # no matter the storage type if size is not going to be exceeded for either, then item can just be added to cache
            else:
                return_val = func(instance, *args, **kwargs)
                self.cache.loc[len(self.cache.index)] = [
                    instance._series["uuid"],
                    func.__name__,
                    args,
                    kwargs,
                    return_val,
                    time.time(),
                ]

            return _return_wrapper(return_val, copy_bool=return_copy)

        return _use_cache

    def invalidate(self, pre: bool = True, post: bool = True):
        """
        invalidate all cache entries associated to a single batch item

        Parameters
        ----------
        pre: bool
            invalidate before the decorated function has been fun

        post: bool
            invalidate after the decorated function has been fun

        """
        def _invalidate(func):
            @wraps(func)
            def __invalidate(instance, *args, **kwargs):
                u = instance._series["uuid"]

                if pre:
                    self.cache.drop(
                        self.cache.loc[self.cache["uuid"] == u].index,
                        inplace=True
                    )

                rval = func(instance, *args, **kwargs)

                if post:
                    self.cache.drop(
                        self.cache.loc[self.cache["uuid"] == u].index,
                        inplace=True
                    )

                return rval

            return __invalidate
        return _invalidate



from .batch_utils import (
    set_parent_raw_data_path,
    get_parent_raw_data_path,
    load_batch,
    create_batch,
)
from .caiman_extensions import *
from pathlib import Path


with open(Path(__file__).parent.joinpath("VERSION"), "r") as f:
    __version__ = f.read().split("\n")[0]

__all__ = [
    "set_parent_raw_data_path",
    "get_parent_raw_data_path",
    "load_batch",
    "create_batch",
    "CaimanDataFrameExtensions",
    "CaimanSeriesExtensions",
    "CNMFExtensions",
    "MCorrExtensions",
]



from pathlib import Path
from typing import *
import numpy as np
import pandas as pd
from caiman import load_memmap
from caiman.source_extraction.cnmf import CNMF
from caiman.source_extraction.cnmf.cnmf import load_CNMF
from caiman.utils.visualization import get_contours as caiman_get_contours
from functools import wraps
import os
from copy import deepcopy

from ._utils import validate
from .cache import Cache
from ..arrays import *
from ..arrays._base import LazyArray


cnmf_cache = Cache()


# this decorator MUST be called BEFORE caching decorators!
def _component_indices_parser(func):
    @wraps(func)
    def _parser(instance, *args, **kwargs) -> Any:
        if "component_indices" in kwargs.keys():
            component_indices: Union[np.ndarray, str, None] = kwargs["component_indices"]
        elif len(args) > 0:
            component_indices = args[0]  # always first positional arg in the extensions
        else:
            component_indices = None  # default

        cnmf_obj = instance.get_output()

        # TODO: finally time to learn Python's new switch case
        accepted = (np.ndarray, str, type(None))
        if not isinstance(component_indices, accepted):
            raise TypeError(f"`component_indices` must be one of type: {accepted}")

        if isinstance(component_indices, np.ndarray):
            pass

        elif component_indices is None:
            component_indices = np.arange(cnmf_obj.estimates.A.shape[1])

        if isinstance(component_indices, str):
            accepted = ["all", "good", "bad"]
            if component_indices not in accepted:
                raise ValueError(f"Accepted `str` values for `component_indices` are: {accepted}")

            if component_indices == "all":
                component_indices = np.arange(cnmf_obj.estimates.A.shape[1])

            elif component_indices == "good":
                component_indices = cnmf_obj.estimates.idx_components

            elif component_indices == "bad":
                component_indices = cnmf_obj.estimates.idx_components_bad
        if "component_indices" in kwargs.keys():
            kwargs["component_indices"] = component_indices
        else:
            args = (component_indices, *args[1:])

        return func(instance, *args, **kwargs)
    return _parser


def _check_permissions(func):
    @wraps(func)
    def __check(instance, *args, **kwargs):
        cnmf_obj_path = instance.get_output_path()

        if not os.access(cnmf_obj_path, os.W_OK):
            raise PermissionError(
                "You do not have write access to the hdf5 output file for this batch item"
            )

        return func(instance, *args, **kwargs)
    return __check


@pd.api.extensions.register_series_accessor("cnmf")
class CNMFExtensions:
    """
    Extensions for managing CNMF output data
    """

    def __init__(self, s: pd.Series):
        self._series = s

    @validate("cnmf")
    def get_cnmf_memmap(self, mode: str = "r") -> np.ndarray:
        """
        Get the CNMF C-order memmap. This should NOT be used for viewing the
        movie frames use ``caiman.get_input_movie()`` for that purpose.

        Parameters
        ----------

        mode: str
            passed to numpy.memmap

            one of: `{'r+', 'r', 'w+', 'c'}`

            The file is opened in this mode:

            +------+-------------------------------------------------------------+
            | 'r'  | Open existing file for reading only.                        |
            +------+-------------------------------------------------------------+
            | 'r+' | Open existing file for reading and writing.                 |
            +------+-------------------------------------------------------------+
            | 'w+' | Create or overwrite existing file for reading and writing.  |
            +------+-------------------------------------------------------------+
            | 'c'  | Copy-on-write: assignments affect data in memory, but       |
            |      | changes are not saved to disk.  The file on disk is         |
            |      | read-only.                                                  |
            +------+-------------------------------------------------------------+

        Returns
        -------
        np.ndarray
            numpy memmap array used for CNMF
        """

        path = self._series.paths.resolve(self._series["outputs"]["cnmf-memmap-path"])
        # Get order f images
        Yr, dims, T = load_memmap(str(path), mode=mode)
        images = np.reshape(Yr.T, [T] + list(dims), order="F")
        return images

    @validate("cnmf")
    def get_output_path(self) -> Path:
        """
        Get the path to the cnmf hdf5 output file.

        **Note:** You generally want to work with the other extensions instead of directly using the hdf5 file.

        Returns
        -------
        Path
            full path to the caiman-format CNMF hdf5 output file

        """

        return self._series.paths.resolve(self._series["outputs"]["cnmf-hdf5-path"])

    @validate("cnmf")
    @cnmf_cache.use_cache
    def get_output(self, return_copy=True) -> CNMF:
        """
        Parameters
        ----------
        return_copy: bool
            | if ``True`` returns a copy of the cached value in memory.
            | if ``False`` returns the same object as the cached value in memory, not recommend this could result in strange unexpected behavior.
            | In general you want a copy of the cached value.

        Returns
        -------
        CNMF
            Returns the Caiman CNMF object

        Examples
        --------

        Load the CNMF model with estimates from the hdf5 file.

        .. code-block:: python

            from mesmerize_core import load_batch

            df = load_batch("/path/to/batch_dataframe_file.pickle")

            # assume the 0th index is a cnmf item
            cnmf_obj = df.iloc[0].cnmf.get_output()

            # see some estimates
            print(cnmf_obj.estimates.C)
            print(cnmf_obj.estimates.f)

        """

        # Need to create a cache object that takes the item's UUID and returns based on that
        # collective global cache
        return load_CNMF(self.get_output_path())

    @validate("cnmf")
    @_component_indices_parser
    @cnmf_cache.use_cache
    def get_masks(
        self, component_indices: Union[np.ndarray, str] = None, threshold: float = 0.01, return_copy=True
    ) -> np.ndarray:
        """
        | Get binary masks of the spatial components at the given ``component_indices``.
        | Created from ``CNMF.estimates.A``

        Parameters
        ----------
        component_indices: str or np.ndarray, optional
            | indices of the components to include
            | if not provided, ``None``, or ``"all"`` uses all components
            | if ``"good"`` uses good components, i.e. ``Estimates.idx_components``
            | if ``"bad"`` uses bad components, i.e. ``Estimates.idx_components_bad``
            | if ``np.ndarray``, uses the indices in the provided array

        threshold: float
            threshold

        return_copy: bool
            | if ``True`` returns a copy of the cached value in memory.
            | if ``False`` returns the same object as the cached value in memory, not recommend this could result in strange unexpected behavior.
            | In general you want a copy of the cached value.

        Returns
        -------
        np.ndarray
            shape is [dim_0, dim_1, n_components]

        """

        cnmf_obj = self.get_output()

        dims = cnmf_obj.dims
        if dims is None:
            dims = cnmf_obj.estimates.dims

        masks = np.zeros(shape=(dims[0], dims[1], len(component_indices)), dtype=bool)

        for n, ix in enumerate(component_indices):
            s = cnmf_obj.estimates.A[:, ix].toarray().reshape(cnmf_obj.dims)
            s[s >= threshold] = 1
            s[s < threshold] = 0

            masks[:, :, n] = s.astype(bool)

        return masks

    @staticmethod
    def _get_spatial_contours(
        cnmf_obj: CNMF, component_indices, swap_dim
    ):

        dims = cnmf_obj.dims
        if dims is None:
            # I think that one of these is `None` if loaded from an hdf5 file
            dims = cnmf_obj.estimates.dims

        # need to transpose these
        if swap_dim:
            dims = dims[1], dims[0]
        else:
            dims = dims[0], dims[1]

        contours = caiman_get_contours(
            cnmf_obj.estimates.A[:, component_indices], dims, swap_dim=swap_dim
        )

        return contours

    @validate("cnmf")
    @_component_indices_parser
    @cnmf_cache.use_cache
    def get_contours(
            self,
            component_indices: Union[np.ndarray, str] = None,
            swap_dim: bool = True,
            return_copy=True
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """
        Get the contour and center of mass for each spatial footprint

        Parameters
        ----------
        component_indices: str or np.ndarray, optional
            | indices of the components to include
            | if not provided, ``None``, or ``"all"`` uses all components
            | if ``"good"`` uses good components, i.e. ``Estimates.idx_components``
            | if ``"bad"`` uses bad components, i.e. ``Estimates.idx_components_bad``
            | if ``np.ndarray``, uses the indices in the provided array

        swap_dim: bool
            swap the x and y coordinates, use if the contours don't align with the cells in your image

        return_copy: bool
            | if ``True`` returns a copy of the cached value in memory.
            | if ``False`` returns the same object as the cached value in memory, not recommend this could result in strange unexpected behavior.
            | In general you want a copy of the cached value.

        Returns
        -------
        Tuple[List[np.ndarray], List[np.ndarray]]
            | (List[coordinates array], List[centers of masses array])
            | each array of coordinates is 2D, [xs, ys]
            | each center of mass is [x, y]

        """

        cnmf_obj = self.get_output()
        contours = self._get_spatial_contours(cnmf_obj, component_indices, swap_dim)

        coordinates = list()
        coms = list()

        for contour in contours:
            coors = contour["coordinates"]
            coors = coors[~np.isnan(coors).any(axis=1)]
            coordinates.append(coors)

            com = coors.mean(axis=0)
            coms.append(com)

        return coordinates, coms

    @validate("cnmf")
    @_component_indices_parser
    @cnmf_cache.use_cache
    def get_temporal(
        self,
        component_indices: Union[np.ndarray, str] = None,
        add_background: bool = False,
        add_residuals: bool = False,
        return_copy=True
    ) -> np.ndarray:
        """
        Get the temporal components for this CNMF item, basically ``CNMF.estimates.C``

        Parameters
        ----------
        component_indices: str or np.ndarray, optional
            | indices of the components to include
            | if not provided, ``None``, or ``"all"`` uses all components
            | if ``"good"`` uses good components, i.e. ``Estimates.idx_components``
            | if ``"bad"`` uses bad components, i.e. ``Estimates.idx_components_bad``
            | if ``np.ndarray``, uses the indices in the provided array

        add_background: bool, default False
            if ``True``, add the temporal background, adds ``cnmf.estimates.f``

        add_residuals: bool, default False
            if ``True``, add residuals, i.e. ``cnmf.estimates.YrA``

        return_copy: bool
            | if ``True`` returns a copy of the cached value in memory.
            | if ``False`` returns the same object as the cached value in memory, not recommend this could result in strange unexpected behavior.
            | In general you want a copy of the cached value.

        Returns
        -------
        np.ndarray
            shape is [n_components, n_frames]

        Examples
        --------

        Plot the temporal components as a heatmap

        .. code-block:: python

            from mesmerize_core import load_batch
            from fastplotlib import Plot

            df = load_batch("/path/to/batch_dataframe_file.pickle")

            # assumes 0th index is a cnmf batch item
            temporal = df.iloc[0].cnmf.get_temporal()

            plot = Plot()

            plot.add_line_collection(temporal)

            plot.show()
        """

        cnmf_obj = self.get_output()

        C = cnmf_obj.estimates.C[component_indices]
        f = cnmf_obj.estimates.f

        temporal = C

        if add_background:
            temporal += f
        elif add_residuals:
            temporal += cnmf_obj.estimates.YrA

        return temporal

    @validate("cnmf")
    @_component_indices_parser
    @cnmf_cache.use_cache
    def get_rcm(
            self,
            component_indices: Union[np.ndarray, str] = None,
            temporal_components: np.ndarray = None,
            return_copy=False
    ) -> LazyArrayRCM:
        """
        Return the reconstructed movie with no background, i.e. ``A ⊗ C``, as a ``LazyArray``.
        This is an array that performs lazy computation of the reconstructed movie only upon indexing.

        Parameters
        ----------
        component_indices: optional, Union[np.ndarray, str]
            | indices of the components to include
            | if ``np.ndarray``, uses these indices in the provided array
            | if ``"good"`` uses good components, i.e. cnmf.estimates.idx_components
            | if ``"bad"`` uses bad components, i.e. cnmf.estimates.idx_components_bad
            | if not provided, ``None``, or ``"all"`` uses all components

        temporal_components: optional, np.ndarray
            temporal components to use as ``C`` for computing reconstructed movie.

            | uses ``cnmf.estimates.C`` if not provided
            | useful if you want to create the reconstructed movie using dF/Fo, z-scored data, etc.

        return_copy: bool, default ``False``
            | if ``True`` returns a copy of the cached value in memory.
            | if ``False`` returns the same object as the cached value in memory
            | ``False`` is used by default when returning ``LazyArrays`` for technical reasons

        Returns
        -------
        LazyArrayRCM
            shape is [n_frames, x_dims, y_dims]

        Examples
        --------

        This example uses fastplotlib to display the reconstructed movie from a CNMF item that has already been run.

        | **fastplotlib code must be run in a notebook**

        | See the demo notebooks for more detailed examples.

        .. code-block:: python

            from mesmerize_core import *
            from fastplotlib.widgets import ImageWidget

            # load existing batch
            df = load_batch("/path/to/batch.pickle")

            # get the reconstructed movie as LazyArray
            # assumes the last index, `-1`, is a cnmf item
            # uses only the "good" components
            rcm = df.iloc[-1].cnmf.get_rcm(component_indices="good")

            # view with ImageWidget
            iw = ImageWidget(data=rcm)
            iw.show()
        """

        cnmf_obj = self.get_output()

        if temporal_components is None:
            temporal_components = cnmf_obj.estimates.C

        else:  # number of spatial components must equal number of temporal components
            if cnmf_obj.estimates.A.shape[1] != temporal_components.shape[0]:
                raise ValueError(
                    f"Number of temporal components provided: `{temporal_components.shape[0]}` "
                    f"does not equal number of spatial components provided: `{cnmf_obj.estimates.A.shape[1]}`"
                )

        if cnmf_obj.estimates.dims is not None:
            dims = cnmf_obj.estimates.dims
        elif cnmf_obj.dims is not None:
            dims = cnmf_obj.dims
        else:
            raise AttributeError(f"`dims` not found in the CNMF data, it is usually found in one of the following:\n"
                                 f"`cnmf_obj.estimates.dims` or `cnmf_obj.dims`")

        spatial = cnmf_obj.estimates.A[:, component_indices]
        temporal = temporal_components[component_indices, :]

        return LazyArrayRCM(spatial=spatial, temporal=temporal, frame_dims=dims)

    @validate("cnmf")
    @cnmf_cache.use_cache
    def get_rcb(self,) -> LazyArrayRCB:
        """
        Return the reconstructed background, ``(b ⊗ f)``

        Returns
        -------
        LazyArrayRCB
            shape is [n_frames, x_dims, y_dims]

        Examples
        --------

        This example uses fastplotlib to display the reconstructed movie from a CNMF item that has already been run.

        | **fastplotlib code must be run in a notebook**

        | See the demo notebooks for more detailed examples.

        .. code-block:: python

            from mesmerize_core import *
            from fastplotlib.widgets import ImageWidget

            # load existing batch
            df = load_batch("/path/to/batch.pickle")

            # get the reconstructed background as a LazyArray
            # assumes the last index, `-1`, is a cnmf item
            rcb = df.iloc[-1].cnmf.get_rcb()

            # view with ImageWidget
            iw = ImageWidget(data=rcb)
            iw.show()
        """

        cnmf_obj = self.get_output()

        if cnmf_obj.estimates.dims is not None:
            dims = cnmf_obj.estimates.dims
        elif cnmf_obj.dims is not None:
            dims = cnmf_obj.dims
        else:
            raise AttributeError(f"`dims` not found in the CNMF data, it is usually found in one of the following:\n"
                                 f"`cnmf_obj.estimates.dims` or `cnmf_obj.dims`")

        spatial = cnmf_obj.estimates.b
        temporal = cnmf_obj.estimates.f

        return LazyArrayRCB(spatial=spatial, temporal=temporal, frame_dims=dims)

    @validate("cnmf")
    @cnmf_cache.use_cache
    def get_residuals(self) -> LazyArrayResiduals:
        """
        Return residuals, ``Y - (A ⊗ C) - (b ⊗ f)``

        Returns
        -------
        LazyArrayResiduals
            shape is [n_frames, x_dims, y_dims]

        Examples
        --------

        This example uses fastplotlib to display the reconstructed movie from a CNMF item that has already been run.

        | **fastplotlib code must be run in a notebook**

        | See the demo notebooks for more detailed examples.

        .. code-block:: python

            from mesmerize_core import *
            from fastplotlib.widgets import ImageWidget

            # load existing batch
            df = load_batch("/path/to/batch.pickle")

            # get the reconstructed background as a LazyArray
            # assumes the last index, `-1`, is a cnmf item
            residuals = df.iloc[-1].cnmf.get_residuals()

            # view with ImageWidget
            iw = ImageWidget(data=residuals)
            iw.show()
        """

        residuals = LazyArrayResiduals(
            self._series.caiman.get_input_movie(),
            self.get_rcm(),
            self.get_rcb(),
        )

        return residuals

    @validate("cnmf")
    @_check_permissions
    @cnmf_cache.invalidate()
    def run_detrend_dfof(
            self,
            quantileMin: float = 8,
            frames_window: int = 500,
            flag_auto: bool = True,
            use_fast: bool = False,
            use_residuals: bool = True,
            detrend_only: bool = False
    ) -> None:
        """
        | Uses caiman's detrend_df_f.
        | call ``CNMF.get_detrend_dfof()`` to get the values.
        | Sets ``CNMF.estimates.F_dff``

        Warnings
        --------
        Overwrites the existing cnmf hdf5 output file for this batch item

        Parameters
        ----------
        quantileMin: float
            quantile used to estimate the baseline (values in [0,100])
            used only if 'flag_auto' is False, i.e. ignored by default

        frames_window: int
            number of frames for computing running quantile

        flag_auto: bool
            flag for determining quantile automatically

        use_fast: bool
            flag for using approximate fast percentile filtering

        detrend_only: bool
            flag for only subtracting baseline and not normalizing by it.
            Used in 1p data processing where baseline fluorescence cannot be
            determined.

        Returns
        -------
        None

        Notes
        ------
        invalidates the cache for this batch item.

        """

        cnmf_obj: CNMF = self.get_output()
        cnmf_obj.estimates.detrend_df_f(
            quantileMin=quantileMin,
            frames_window=frames_window,
            flag_auto=flag_auto,
            use_fast=use_fast,
            use_residuals=use_residuals,
            detrend_only=detrend_only
        )

        # remove current hdf5 file
        cnmf_obj_path = self.get_output_path()
        cnmf_obj_path.unlink()

        # save new hdf5 file with new F_dff vals
        cnmf_obj.save(str(cnmf_obj_path))

    @validate("cnmf")
    @_component_indices_parser
    @cnmf_cache.use_cache
    def get_detrend_dfof(
            self,
            component_indices: Union[np.ndarray, str] = None,
            return_copy: bool = True
    ):
        """
        Get the detrended dF/F0 curves after calling ``run_detrend_dfof``.
        Basically ``CNMF.estimates.F_dff``.

        Parameters
        ----------
        component_indices: str or np.ndarray, optional
            | indices of the components to include
            | if not provided, ``None``, or ``"all"`` uses all components
            | if ``"good"`` uses good components, i.e. ``Estimates.idx_components``
            | if ``"bad"`` uses bad components, i.e. ``Estimates.idx_components_bad``
            | if ``np.ndarray``, uses the indices in the provided array

        return_copy: bool
            | if ``True`` returns a copy of the cached value in memory.
            | if ``False`` returns the same object as the cached value in memory, not recommend this could result in strange unexpected behavior.
            | In general you want a copy of the cached value.

        Returns
        -------
        np.ndarray
            shape is [n_components, n_frames]

        """

        cnmf_obj = self.get_output()
        if cnmf_obj.estimates.F_dff is None:
            raise AttributeError("You must run ``cnmf.run_detrend_dfof()`` first")

        return cnmf_obj.estimates.F_dff[component_indices]

    @validate("cnmf")
    @_check_permissions
    @cnmf_cache.invalidate()
    def run_eval(self, params: dict) -> None:
        """
        Run component evaluation. This basically changes the indices for good and bad components.

        Warnings
        --------
        Overwrites the existing cnmf hdf5 output file for this batch item

        Parameters
        ----------
        params: dict
            dict of parameters for component evaluation

            ==============  =================
            parameter       details
            ==============  =================
            SNR_lowest      ``float``, minimum accepted SNR value
            cnn_lowest      ``float``, minimum accepted value for CNN classifier
            gSig_range      ``List[int, int]`` or ``None``, range for gSig scale for CNN classifier
            min_SNR         ``float``, transient SNR threshold
            min_cnn_thr     ``float``, threshold for CNN classifier
            rval_lowest     ``float``, minimum accepted space correlation
            rval_thr        ``float``, space correlation threshold
            use_cnn         ``bool``, use CNN based classifier
            use_ecc         ``bool``, flag for eccentricity based filtering
            max_ecc         ``float``, max eccentricity
            ==============  =================

        Returns
        -------
        None

        Notes
        ------
        invalidates the cache for this batch item.

        """

        cnmf_obj = self.get_output()

        valid = list(cnmf_obj.params.quality.keys())
        for k in params.keys():
            if k not in valid:
                raise KeyError(
                    f"passed params dict key `{k}` is not a valid parameter for quality evaluation\n"
                    f"valid param keys are: {valid}"
                )

        cnmf_obj.params.quality.update(params)
        cnmf_obj.estimates.filter_components(
            imgs=self._series.caiman.get_input_movie(),
            params=cnmf_obj.params
        )

        cnmf_obj_path = self.get_output_path()
        cnmf_obj_path.unlink()

        cnmf_obj.save(str(cnmf_obj_path))
        self._series["params"]["eval"] = deepcopy(params)

    @validate("cnmf")
    def get_good_components(self) -> np.ndarray:
        """
        get the good component indices, ``Estimates.idx_components``

        Returns
        -------
        np.ndarray
            array of ints, indices of good components

        """

        cnmf_obj = self.get_output()
        return cnmf_obj.estimates.idx_components

    @validate("cnmf")
    def get_bad_components(self) -> np.ndarray:
        """
        get the bad component indices, ``Estimates.idx_components_bad``

        Returns
        -------
        np.ndarray
            array of ints, indices of bad components

        """

        cnmf_obj = self.get_output()
        return cnmf_obj.estimates.idx_components_bad



import os

import numpy as np
from caiman.utils.utils import load_dict_from_hdf5
from caiman.source_extraction.cnmf import cnmf
import numpy.testing
import pandas as pd
from mesmerize_core import (
    create_batch,
    load_batch,
    CaimanDataFrameExtensions,
    CaimanSeriesExtensions,
    set_parent_raw_data_path,
)
from mesmerize_core.batch_utils import DATAFRAME_COLUMNS, COMPUTE_BACKEND_SUBPROCESS, get_full_raw_data_path
from mesmerize_core.utils import IS_WINDOWS
from uuid import uuid4
from typing import *
import pytest
import requests
from tqdm import tqdm
from .params import test_params
from uuid import UUID
from pathlib import Path
import shutil
from zipfile import ZipFile
from pprint import pprint
from mesmerize_core.caiman_extensions import cnmf
import time
import tifffile
from copy import deepcopy

tmp_dir = Path(os.path.dirname(os.path.abspath(__file__)), "tmp")
vid_dir = Path(os.path.dirname(os.path.abspath(__file__)), "videos")
ground_truths_dir = Path(os.path.dirname(os.path.abspath(__file__)), "ground_truths")
ground_truths_file = Path(
    os.path.dirname(os.path.abspath(__file__)), "ground_truths.zip"
)

os.makedirs(tmp_dir, exist_ok=True)
os.makedirs(vid_dir, exist_ok=True)
os.makedirs(ground_truths_dir, exist_ok=True)


def _download_ground_truths():
    print(f"Downloading ground truths")
    url = f"https://zenodo.org/record/6828096/files/ground_truths.zip"

    # basically from https://stackoverflow.com/questions/37573483/progress-bar-while-download-file-over-http-with-requests/37573701
    response = requests.get(url, stream=True)
    total_size_in_bytes = int(response.headers.get("content-length", 0))
    block_size = 1024  # 1 Kibibyte
    progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)

    with open(ground_truths_file, "wb") as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()

    ZipFile(ground_truths_file).extractall(ground_truths_dir.parent)


if len(list(ground_truths_dir.iterdir())) == 0:
    _download_ground_truths()

elif "DOWNLOAD_GROUND_TRUTHS" in os.environ.keys():
    if os.environ["DOWNLOAD_GROUND_TRUTHS"] == "1":
        _download_ground_truths()


def get_tmp_filename():
    return os.path.join(tmp_dir, f"{uuid4()}.pickle")


def clear_tmp():
    if "MESMERIZE_KEEP_TEST_DATA" in os.environ.keys():
        if os.environ["MESMERIZE_KEEP_TEST_DATA"] == "1":
            return

    shutil.rmtree(tmp_dir)
    shutil.rmtree(vid_dir)


def get_datafile(fname: str):
    local_path = Path(os.path.join(vid_dir, f"{fname}.tif"))
    if local_path.is_file():
        return local_path
    else:
        return download_data(fname)


def download_data(fname: str):
    """
    Download the large network files from Zenodo
    """
    url = {
        "mcorr": "https://caiman.flatironinstitute.org/~neuro/caiman_downloadables/demoMovie.tif",
        "cnmf": None,
        "cnmfe": "https://caiman.flatironinstitute.org/~neuro/caiman_downloadables/data_endoscope.tif",
    }.get(fname)

    print(f"Downloading test data from: {url}")

    response = requests.get(url, stream=True)
    total_size_in_bytes = int(response.headers.get("content-length", 0))
    block_size = 1024  # 1 Kibibyte
    progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)
    path = os.path.join(vid_dir, f"{fname}.tif")

    with open(path, "wb") as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)

    progress_bar.close()

    if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
        raise ConnectionError("Couldn't download test test")

    return Path(path)


def teardown_module():
    if IS_WINDOWS:  # because windows is stupid with permissions
        return
    clear_tmp()


def _create_tmp_batch() -> Tuple[pd.DataFrame, str]:
    fname = get_tmp_filename()
    df = create_batch(fname)

    return df, fname


def test_create_batch():
    df, fname = _create_tmp_batch()

    for c in DATAFRAME_COLUMNS:
        assert c in df.columns

    # test that existing batch is not overwritten
    with pytest.raises(FileExistsError):
        create_batch(fname)


def test_mcorr():
    set_parent_raw_data_path(vid_dir)
    algo = "mcorr"
    df, batch_path = _create_tmp_batch()

    batch_path = Path(batch_path)
    batch_dir = batch_path.parent

    print(f"Testing mcorr")
    input_movie_path = get_datafile(algo)
    print(input_movie_path)
    df.caiman.add_item(
        algo=algo,
        item_name=f"test-{algo}",
        input_movie_path=input_movie_path,
        params=test_params[algo],
    )

    assert df.iloc[-1]["algo"] == algo
    assert df.iloc[-1]["item_name"] == f"test-{algo}"
    assert df.iloc[-1]["params"] == test_params[algo]
    assert df.iloc[-1]["outputs"] is None
    try:
        UUID(df.iloc[-1]["uuid"])
    except:
        pytest.fail("Something wrong with setting UUID for batch items")

    # test that batch path is propagated to pd.Series
    assert (
            df.attrs["batch_path"]
            == df.paths.get_batch_path()
            == df.iloc[-1].paths.get_batch_path()
            == df.iloc[-1].attrs["batch_path"]
    )

    # test that path resolve works for parent_raw_dir
    rel_input_movie_path = input_movie_path.relative_to(vid_dir)
    assert (
            df.paths.resolve(rel_input_movie_path)
            == df.iloc[-1].paths.resolve(rel_input_movie_path)
            == input_movie_path
    )
    # test that path splitting works for parent_raw_dir
    split = (vid_dir, input_movie_path.relative_to(vid_dir))
    assert (
            df.paths.split(input_movie_path)
            == df.iloc[-1].paths.split(input_movie_path)
            == split
    )
    # test that the input_movie_path in the DataFrame rows are relative
    assert Path(df.iloc[-1]["input_movie_path"]) == split[1]

    assert (
            get_full_raw_data_path(df.iloc[-1]["input_movie_path"])
            == vid_dir.joinpath(f"{algo}.tif")
            == vid_dir.joinpath(df.iloc[-1]["input_movie_path"])
            == df.paths.resolve(df.iloc[-1]["input_movie_path"])
    )

    process = df.iloc[-1].caiman.run()
    # process.wait()

    df = load_batch(batch_path)

    with pd.option_context("display.max_rows", None, "display.max_columns", None):
        print(df)

    pprint(df.iloc[-1]["outputs"], width=-1)
    print(df.iloc[-1]["outputs"]["traceback"])
    assert df.iloc[-1]["outputs"]["success"] is True
    assert df.iloc[-1]["outputs"]["traceback"] is None

    # test that path resolve works for batch_dir
    mcorr_memmap_path = batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]),
        f'{df.iloc[-1]["uuid"]}-mcorr_els__d1_60_d2_80_d3_1_order_F_frames_2000.mmap',
    )
    rel_mcorr_memmap_path = mcorr_memmap_path.relative_to(batch_dir)
    assert (
            df.paths.resolve(rel_mcorr_memmap_path)
            == df.iloc[-1].paths.resolve(rel_mcorr_memmap_path)
            == mcorr_memmap_path
    )
    # test that path splitting works for batch_dir
    split = (batch_dir, mcorr_memmap_path.relative_to(batch_dir))
    assert (
            df.paths.split(mcorr_memmap_path)
            == df.iloc[-1].paths.split(mcorr_memmap_path)
            == split
    )

    assert (
            input_movie_path
            == df.iloc[-1].caiman.get_input_movie_path()
            == df.paths.resolve(df.iloc[-1]["input_movie_path"])
    )

    # test to check mmap output path
    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["mcorr-output-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["mcorr-output-path"])
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]),
        f'{df.iloc[-1]["uuid"]}-mcorr_els__d1_60_d2_80_d3_1_order_F_frames_2000.mmap',
    )
    )

    # test to check mean-projection output path
    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["mean-projection-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["mean-projection-path"])
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}_mean_projection.npy'
    )
    )

    # test to check std-projection output path
    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["std-projection-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["std-projection-path"])
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}_std_projection.npy'
    )
    )

    # test to check max-projection output path
    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["max-projection-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["max-projection-path"])
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}_max_projection.npy'
    )
    )

    # test to check correlation image output path
    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["corr-img-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["corr-img-path"])
            == batch_dir.joinpath(str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}_cn.npy')
    )

    # test to check mcorr get_output_path()
    assert (
            df.iloc[-1].mcorr.get_output_path()
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]),
        f'{df.iloc[-1]["uuid"]}-mcorr_els__d1_60_d2_80_d3_1_order_F_frames_2000.mmap',
    )
            == df.paths.resolve(df.iloc[-1]["outputs"]["mcorr-output-path"])
    )

    # test to check mcorr get_output()
    mcorr_output = df.iloc[-1].mcorr.get_output()
    mcorr_output_actual = numpy.load(
        ground_truths_dir.joinpath("mcorr", "mcorr_output.npy")
    )
    numpy.testing.assert_array_equal(mcorr_output, mcorr_output_actual)

    # test to check caiman get_input_movie_path()
    assert df.iloc[-1].caiman.get_input_movie_path() == get_full_raw_data_path(
        df.iloc[0]["input_movie_path"]
    )

    # test to check caiman get_correlation_img()
    mcorr_corr_img = df.iloc[-1].caiman.get_corr_image()
    mcorr_corr_img_actual = numpy.load(
        ground_truths_dir.joinpath("mcorr", "mcorr_correlation_img.npy")
    )
    numpy.testing.assert_allclose(mcorr_corr_img, mcorr_corr_img_actual, rtol=1e-2, atol=1e-10)

    # test to check caiman get_projection("mean")
    mcorr_mean = df.iloc[-1].caiman.get_projection("mean")
    mcorr_mean_actual = numpy.load(
        ground_truths_dir.joinpath("mcorr", "mcorr_mean.npy")
    )
    numpy.testing.assert_array_equal(mcorr_mean, mcorr_mean_actual)

    # test to check caiman get_projection("std")
    mcorr_std = df.iloc[-1].caiman.get_projection("std")
    mcorr_std_actual = numpy.load(ground_truths_dir.joinpath("mcorr", "mcorr_std.npy"))
    numpy.testing.assert_array_equal(mcorr_std, mcorr_std_actual)

    # test to check caiman get_projection("max")
    mcorr_max = df.iloc[-1].caiman.get_projection("max")
    mcorr_max_actual = numpy.load(ground_truths_dir.joinpath("mcorr", "mcorr_max.npy"))
    numpy.testing.assert_array_equal(mcorr_max, mcorr_max_actual)


def test_cnmf():
    set_parent_raw_data_path(vid_dir)
    algo = "mcorr"

    df, batch_path = _create_tmp_batch()

    batch_path = Path(batch_path)
    batch_dir = batch_path.parent

    input_movie_path = get_datafile(algo)
    print(input_movie_path)

    df.caiman.add_item(
        algo=algo,
        item_name=f"test-{algo}",
        input_movie_path=input_movie_path,
        params=test_params[algo],
    )

    assert df.iloc[-1]["algo"] == algo
    assert df.iloc[-1]["item_name"] == f"test-{algo}"
    assert df.iloc[-1]["params"] == test_params[algo]
    assert df.iloc[-1]["outputs"] is None
    try:
        UUID(df.iloc[-1]["uuid"])
    except:
        pytest.fail("Something wrong with setting UUID for batch items")

    assert vid_dir.joinpath(df.iloc[-1]["input_movie_path"]) == vid_dir.joinpath(
        f"{algo}.tif"
    )

    process = df.iloc[-1].caiman.run()
    # process.wait()

    df = load_batch(batch_path)

    with pd.option_context("display.max_rows", None, "display.max_columns", None):
        print(df)

    pprint(df.iloc[-1]["outputs"], width=-1)
    print(df.iloc[-1]["outputs"]["traceback"])
    assert df.iloc[-1]["outputs"]["success"] is True
    assert df.iloc[-1]["outputs"]["traceback"] is None

    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["mcorr-output-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["mcorr-output-path"])
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]),
        f'{df.iloc[-1]["uuid"]}-mcorr_els__d1_60_d2_80_d3_1_order_F_frames_2000.mmap',
    )
    )

    algo = "cnmf"
    print("Testing cnmf")
    input_movie_path = df.iloc[-1].mcorr.get_output_path()
    df.caiman.add_item(
        algo=algo,
        item_name=f"test-{algo}",
        input_movie_path=input_movie_path,
        params=test_params[algo],
    )

    assert df.iloc[-1]["algo"] == algo
    assert df.iloc[-1]["item_name"] == f"test-{algo}"
    assert df.iloc[-1]["params"] == test_params[algo]
    assert df.iloc[-1]["outputs"] is None
    try:
        UUID(df.iloc[-1]["uuid"])
    except:
        pytest.fail("Something wrong with setting UUID for batch items")
    print("cnmf input_movie_path:", df.iloc[-1]["input_movie_path"])
    assert batch_dir.joinpath(df.iloc[-1]["input_movie_path"]) == input_movie_path

    process = df.iloc[-1].caiman.run()
    # process.wait()

    df = load_batch(batch_path)

    with pd.option_context("display.max_rows", None, "display.max_columns", None):
        print(df)

    pprint(df.iloc[-1]["outputs"], width=-1)
    print(df.iloc[-1]["outputs"]["traceback"])

    # Confirm output path is as expected
    assert df.iloc[-1]["outputs"]["success"] is True
    assert df.iloc[-1]["outputs"]["traceback"] is None

    assert (
            input_movie_path
            == df.iloc[-1].caiman.get_input_movie_path()
            == df.paths.resolve(df.iloc[-1]["input_movie_path"])
    )

    assert (
            batch_dir.joinpath(str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}.hdf5')
            == df.paths.resolve(df.iloc[-1]["outputs"]["cnmf-hdf5-path"])
            == batch_dir.joinpath(df.iloc[-1]["outputs"]["cnmf-hdf5-path"])
    )

    # test to check mmap output path
    assert (
            batch_dir.joinpath(
                str(df.iloc[-1]["uuid"]),
                f'{df.iloc[-1]["uuid"]}_cnmf-memmap_d1_60_d2_80_d3_1_order_C_frames_2000.mmap',
            )
            == df.paths.resolve(df.iloc[-1]["outputs"]["cnmf-memmap-path"])
            == batch_dir.joinpath(df.iloc[-1]["outputs"]["cnmf-memmap-path"])
    )

    # test to check mean-projection output path
    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["mean-projection-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["mean-projection-path"])
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}_mean_projection.npy'
    )
    )

    # test to check std-projection output path
    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["std-projection-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["std-projection-path"])
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}_std_projection.npy'
    )
    )

    # test to check max-projection output path
    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["max-projection-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["max-projection-path"])
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}_max_projection.npy'
    )
    )

    # test to check correlation image output path
    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["corr-img-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["corr-img-path"])
            == batch_dir.joinpath(str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}_cn.npy')
    )

    print("testing cnmf.get_cnmf_memmap()")
    # test to check cnmf get_cnmf_memmap()
    cnmf_mmap_output = df.iloc[-1].cnmf.get_cnmf_memmap()
    cnmf_mmap_output_actual = numpy.load(
        ground_truths_dir.joinpath("cnmf", "cnmf_output_mmap.npy")
    )
    numpy.testing.assert_array_equal(cnmf_mmap_output, cnmf_mmap_output_actual)

    print("testing caiman.get_input_movie() for cnmf")
    cnmf_input_mmap = df.iloc[-1].caiman.get_input_movie()
    cnmf_input_mmap_actual = numpy.load(
        ground_truths_dir.joinpath("cnmf", "cnmf_input_mmap.npy")
    )
    numpy.testing.assert_array_equal(cnmf_input_mmap, cnmf_input_mmap_actual)
    # cnmf input memmap from mcorr output should also equal mcorr output
    mcorr_output = df.iloc[-2].mcorr.get_output()
    numpy.testing.assert_array_equal(cnmf_input_mmap, mcorr_output)

    # test to check cnmf get_output_path()
    assert (
            df.iloc[-1].cnmf.get_output_path()
            == batch_dir.joinpath(df.iloc[-1]["outputs"]["cnmf-hdf5-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["cnmf-hdf5-path"])
            == batch_dir.joinpath(str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}.hdf5')
    )

    # test to check cnmf get_output()
    assert isinstance(df.iloc[-1].cnmf.get_output(), cnmf.CNMF)
    # this doesn't work because some keys in the hdf5 file are
    # not always identical, like the path to the mmap file
    # assert sha1(open(df.iloc[1].cnmf.get_output_path(), "rb").read()).hexdigest() == sha1(open(ground_truths_dir.joinpath('cnmf', 'cnmf_output.hdf5'), "rb").read()).hexdigest()

    # test to check cnmf get_masks()
    cnmf_spatial_masks = df.iloc[-1].cnmf.get_masks("good")
    cnmf_spatial_masks_actual = numpy.load(
        ground_truths_dir.joinpath("cnmf", "spatial_masks.npy")
    )
    numpy.testing.assert_array_equal(cnmf_spatial_masks, cnmf_spatial_masks_actual)

    # test to check get_contours()
    cnmf_spatial_contours_contours = df.iloc[-1].cnmf.get_contours("good")[0]
    cnmf_spatial_contours_coms = df.iloc[-1].cnmf.get_contours("good")[1]
    cnmf_spatial_contours_contours_actual = numpy.load(
        ground_truths_dir.joinpath("cnmf", "spatial_contours_contours.npy"),
        allow_pickle=True,
    )
    cnmf_spatial_contours_coms_actual = numpy.load(
        ground_truths_dir.joinpath("cnmf", "spatial_contours_coms.npy"),
        allow_pickle=True,
    )
    for contour, actual_contour in zip(
            cnmf_spatial_contours_contours, cnmf_spatial_contours_contours_actual
    ):
        numpy.testing.assert_allclose(contour, actual_contour, rtol=1e-2, atol=1e-10)
    for com, actual_com in zip(
            cnmf_spatial_contours_coms, cnmf_spatial_contours_coms_actual
    ):
        numpy.testing.assert_allclose(com, actual_com, rtol=1e-2, atol=1e-10)

    # test to check get_temporal()
    cnmf_temporal_components = df.iloc[-1].cnmf.get_temporal("good")
    cnmf_temporal_components_actual = numpy.load(
        ground_truths_dir.joinpath("cnmf", "temporal_components.npy")
    )
    numpy.testing.assert_allclose(
        cnmf_temporal_components, cnmf_temporal_components_actual, rtol=1e-2, atol=1e-10
    )

    # test to check get_rcm()
    cnmf_reconstructed_movie_AouterC = df.iloc[-1].cnmf.get_rcm("all")
    cnmf_reconstructed_movie_AouterC_actual = numpy.load(
        ground_truths_dir.joinpath("cnmf", "reconstructed_movie_new.npy")
    )
    numpy.testing.assert_allclose(
        cnmf_reconstructed_movie_AouterC.as_numpy(),
        cnmf_reconstructed_movie_AouterC_actual, rtol=1e-1, atol=1e-10
    )

    # test that get_item is working properly for LazyArrays
    for i in np.random.randint(10, cnmf_reconstructed_movie_AouterC_actual.shape[0] - 11, size=10):
        numpy.testing.assert_allclose(
            cnmf_reconstructed_movie_AouterC[i],
            cnmf_reconstructed_movie_AouterC_actual[i], rtol=1e-1, atol=1e-10
        )
    for i in np.random.randint(10, cnmf_reconstructed_movie_AouterC_actual.shape[0] - 11, size=10):
        numpy.testing.assert_allclose(
            cnmf_reconstructed_movie_AouterC[i-5:i+5],
            cnmf_reconstructed_movie_AouterC_actual[i-5:i+5], rtol=1e-1, atol=1e-10
        )

    # test to check get_rcb()
    cnmf_reconstructed_background = df.iloc[-1].cnmf.get_rcb()
    cnmf_reconstructed_background_actual = numpy.load(ground_truths_dir.joinpath("cnmf", "reconstructed_background.npy"))
    numpy.testing.assert_allclose(
        cnmf_reconstructed_background.as_numpy(),
        cnmf_reconstructed_background_actual, rtol=1e-2, atol=1e-10
    )

    # test to check get_residuals()
    cnmf_residuals = df.iloc[-1].cnmf.get_residuals()
    # I think something is wrong with the residuals groundtruth file
    # cnmf_residuals_actual = numpy.load(ground_truths_dir.joinpath("cnmf", "residuals.npy"))
    numpy.testing.assert_allclose(
        cnmf_residuals.as_numpy(),
        df.iloc[-1].caiman.get_input_movie() - cnmf_reconstructed_movie_AouterC_actual - cnmf_reconstructed_background_actual,
        rtol=1e2,
        atol=1e-5
    )

    # test to check caiman get_input_movie_path(), should be output of previous mcorr
    assert (
            df.iloc[-1].caiman.get_input_movie_path()
            == df.paths.resolve(df.iloc[-1]["input_movie_path"])
            == batch_dir.joinpath(df.iloc[-1]["input_movie_path"])
    )

    # test to check caiman get_correlation_img()
    cnmf_corr_img = df.iloc[-1].caiman.get_corr_image()
    cnmf_corr_img_actual = numpy.load(
        ground_truths_dir.joinpath("cnmf", "cnmf_correlation_img.npy")
    )
    numpy.testing.assert_allclose(cnmf_corr_img, cnmf_corr_img_actual, rtol=1e-5, atol=1e-5)

    # test to check caiman get_projection("mean")
    cnmf_mean = df.iloc[-1].caiman.get_projection("mean")
    cnmf_mean_actual = numpy.load(ground_truths_dir.joinpath("cnmf", "cnmf_mean.npy"))
    numpy.testing.assert_array_equal(cnmf_mean, cnmf_mean_actual)

    # test to check caiman get_projection("std")
    cnmf_std = df.iloc[-1].caiman.get_projection("std")
    cnmf_std_actual = numpy.load(ground_truths_dir.joinpath("cnmf", "cnmf_std.npy"))
    numpy.testing.assert_array_equal(cnmf_std, cnmf_std_actual)

    # test to check caiman get_projection("max")
    cnmf_max = df.iloc[-1].caiman.get_projection("std")
    cnmf_max_actual = numpy.load(ground_truths_dir.joinpath("cnmf", "cnmf_std.npy"))
    numpy.testing.assert_array_equal(cnmf_max, cnmf_max_actual)

    # test to check passing optional ixs components to various functions
    ixs_components = numpy.array([1, 3, 5, 2])

    # test to check ixs components for cnmf.get_masks()
    ixs_spatial_masks = df.iloc[-1].cnmf.get_masks(ixs_components)
    ixs_spatial_masks_actual = numpy.load(
        ground_truths_dir.joinpath("cnmf", "cnmf_ixs", "ixs_spatial_masks.npy"),
        allow_pickle=True,
    )
    numpy.testing.assert_array_equal(ixs_spatial_masks, ixs_spatial_masks_actual)

    # test to check ixs components for cnmf.get_contours()
    ixs_contours_contours = df.iloc[-1].cnmf.get_contours(ixs_components)[0]
    ixs_contours_contours_actual = numpy.load(
        ground_truths_dir.joinpath(
            "cnmf", "cnmf_ixs", "ixs_spatial_contours_contours.npy"
        ),
        allow_pickle=True,
    )
    ixs_contours_coms = df.iloc[-1].cnmf.get_contours(ixs_components)[1]
    ixs_contours_coms_actual = numpy.load(
        ground_truths_dir.joinpath("cnmf", "cnmf_ixs", "ixs_spatial_contours_coms.npy"),
        allow_pickle=True,
    )
    for contour, actual_contour in zip(
            ixs_contours_contours, ixs_contours_contours_actual
    ):
        numpy.testing.assert_allclose(contour, actual_contour, rtol=1e-2, atol=1e-10)
    for com, actual_com in zip(ixs_contours_coms, ixs_contours_coms_actual):
        numpy.testing.assert_allclose(com, actual_com, rtol=1e-2, atol=1e-10)

    # test to check ixs components for cnmf.get_temporal()
    ixs_temporal_components = df.iloc[-1].cnmf.get_temporal(ixs_components)
    ixs_temporal_components_actual = numpy.load(
        ground_truths_dir.joinpath("cnmf", "cnmf_ixs", "ixs_temporal_components.npy"),
        allow_pickle=True,
    )
    numpy.testing.assert_allclose(
        ixs_temporal_components, ixs_temporal_components_actual, rtol=1e-2, atol=1e-10
    )


def test_cnmfe():
    set_parent_raw_data_path(vid_dir)

    df, batch_path = _create_tmp_batch()

    batch_path = Path(batch_path)
    batch_dir = batch_path.parent

    input_movie_path = get_datafile("cnmfe")
    print(input_movie_path)
    df.caiman.add_item(
        algo="mcorr",
        item_name=f"test-cnmfe-mcorr",
        input_movie_path=input_movie_path,
        params=test_params["mcorr"],
    )
    process = df.iloc[-1].caiman.run()
    # process.wait()

    df = load_batch(batch_path)

    # Test if running full cnmfe works
    print("testing cnmfe")
    algo = "cnmfe"
    param_name = "cnmfe_full"
    input_movie_path = df.iloc[0].mcorr.get_output_path()
    print(input_movie_path)

    df.caiman.add_item(
        algo=algo,
        item_name=f"test-{algo}",
        input_movie_path=input_movie_path,
        params=test_params[param_name],
    )

    assert df.iloc[-1]["algo"] == algo
    assert df.iloc[-1]["item_name"] == f"test-{algo}"
    assert df.iloc[-1]["params"] == test_params[param_name]
    assert df.iloc[-1]["outputs"] is None
    try:
        UUID(df.iloc[-1]["uuid"])
    except:
        pytest.fail("Something wrong with setting UUID for batch items")

    assert (
            batch_dir.joinpath(df.iloc[-1]["input_movie_path"])
            == batch_dir.joinpath(df.iloc[0].mcorr.get_output_path())
            == df.paths.resolve(df.iloc[-1]["input_movie_path"])
    )

    process = df.iloc[-1].caiman.run()
    # process.wait()

    df = load_batch(batch_path)

    with pd.option_context("display.max_rows", None, "display.max_columns", None):
        print(df)

    pprint(df.iloc[-1]["outputs"], width=-1)
    print(df.iloc[-1]["outputs"]["traceback"])

    # Confirm output path is as expected
    assert df.iloc[-1]["outputs"]["success"] is True
    assert df.iloc[-1]["outputs"]["traceback"] is None

    assert (
            input_movie_path
            == df.iloc[-1].caiman.get_input_movie_path()
            == df.paths.resolve(df.iloc[-1]["input_movie_path"])
    )

    assert batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}.hdf5'
    ) == df.paths.resolve(df.iloc[-1]["outputs"]["cnmf-hdf5-path"])

    # test to check mmap output path
    assert (
            batch_dir.joinpath(
                str(df.iloc[-1]["uuid"]),
                f'{df.iloc[-1]["uuid"]}_cnmf-memmap_d1_128_d2_128_d3_1_order_C_frames_1000.mmap',
            )
            == df.paths.resolve(df.iloc[-1]["outputs"]["cnmf-memmap-path"])
            == batch_dir.joinpath(df.iloc[-1]["outputs"]["cnmf-memmap-path"])
    )

    # test to check mean-projection output path
    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["mean-projection-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["mean-projection-path"])
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}_mean_projection.npy'
    )
    )

    # test to check std-projection output path
    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["std-projection-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["std-projection-path"])
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}_std_projection.npy'
    )
    )

    # test to check max-projection output path
    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["max-projection-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["max-projection-path"])
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]), f'{df.iloc[-1]["uuid"]}_max_projection.npy'
    )
    )

    # extension tests - full

    # test to check cnmf get_cnmf_memmap()
    cnmfe_mmap_output = df.iloc[-1].cnmf.get_cnmf_memmap()
    cnmfe_mmap_output_actual = numpy.load(
        ground_truths_dir.joinpath("cnmfe_full", "cnmfe_full_output_mmap.npy")
    )
    numpy.testing.assert_array_equal(cnmfe_mmap_output, cnmfe_mmap_output_actual)

    # test to check input memmap paths
    cnmfe_input_mmap = df.iloc[-1].caiman.get_input_movie()
    cnmfe_input_mmap_actual = numpy.load(
        ground_truths_dir.joinpath("cnmfe_full", "cnmfe_full_input_mmap.npy")
    )
    numpy.testing.assert_array_equal(cnmfe_input_mmap, cnmfe_input_mmap_actual)
    # cnmf input memmap from mcorr output should also equal mcorr output
    mcorr_output = df.iloc[0].mcorr.get_output()
    numpy.testing.assert_array_equal(cnmfe_input_mmap, mcorr_output)

    # test to check cnmf get_output_path()
    assert (
            df.iloc[-1].cnmf.get_output_path()
            == batch_dir.joinpath(df.iloc[-1]["outputs"]["cnmf-hdf5-path"])
            == df.iloc[-1].paths.resolve(df.iloc[-1]["outputs"]["cnmf-hdf5-path"])
    )

    # test to check cnmf get_output()
    assert isinstance(df.iloc[-1].cnmf.get_output(), cnmf.CNMF)
    # this doesn't work because some keys in the hdf5 file are
    # not always identical, like the path to the mmap file
    # assert sha1(open(df.iloc[1].cnmf.get_output_path(), "rb").read()).hexdigest() == sha1(open(ground_truths_dir.joinpath('cnmf', 'cnmf_output.hdf5'), "rb").read()).hexdigest()

    # test to check cnmf get_masks()
    cnmfe_spatial_masks = df.iloc[-1].cnmf.get_masks("good")
    cnmfe_spatial_masks_actual = numpy.load(
        ground_truths_dir.joinpath("cnmfe_full", "cnmfe_spatial_masks.npy")
    )
    numpy.testing.assert_array_equal(cnmfe_spatial_masks, cnmfe_spatial_masks_actual)

    # test to check get_contours()
    cnmfe_spatial_contours_contours = df.iloc[-1].cnmf.get_contours("good")[0]
    cnmfe_spatial_contours_coms = df.iloc[-1].cnmf.get_contours("good")[1]
    cnmfe_spatial_contours_contours_actual = numpy.load(
        ground_truths_dir.joinpath("cnmfe_full", "cnmfe_spatial_contours_contours.npy"),
        allow_pickle=True,
    )
    cnmfe_spatial_contours_coms_actual = numpy.load(
        ground_truths_dir.joinpath("cnmfe_full", "cnmfe_spatial_contours_coms.npy"),
        allow_pickle=True,
    )
    for contour, actual_contour in zip(
            cnmfe_spatial_contours_contours, cnmfe_spatial_contours_contours_actual
    ):
        numpy.testing.assert_allclose(contour, actual_contour, rtol=1e-2, atol=1e-10)
    for com, actual_com in zip(
            cnmfe_spatial_contours_coms, cnmfe_spatial_contours_coms_actual
    ):
        numpy.testing.assert_allclose(com, actual_com, rtol=1e-2, atol=1e-10)

    # test to check get_temporal()
    cnmfe_temporal_components = df.iloc[-1].cnmf.get_temporal("good")
    cnmfe_temporal_components_actual = numpy.load(
        ground_truths_dir.joinpath("cnmfe_full", "cnmfe_temporal_components.npy")
    )
    numpy.testing.assert_allclose(
        cnmfe_temporal_components,
        cnmfe_temporal_components_actual,
        rtol=1e1,
        atol=1e-1,
    )

    # test to check get_rcm()
    cnmfe_reconstructed_movie_AouterC = df.iloc[-1].cnmf.get_rcm("all")
    cnmfe_reconstructed_movie_AouterC_actual = numpy.load(
        ground_truths_dir.joinpath("cnmfe_full", "cnmfe_reconstructed_movie_new.npy")
    )
    numpy.testing.assert_allclose(
        cnmfe_reconstructed_movie_AouterC.as_numpy(),
        cnmfe_reconstructed_movie_AouterC_actual, rtol=1e2, atol=1e-1
    )

    # test to check get_rcb()
    cnmfe_reconstructed_background = df.iloc[-1].cnmf.get_rcb()
    cnmfe_reconstructed_background_actual = numpy.load(
        ground_truths_dir.joinpath("cnmfe_full", "cnmfe_reconstructed_background.npy"))
    numpy.testing.assert_allclose(
        cnmfe_reconstructed_background.as_numpy(),
        cnmfe_reconstructed_background_actual, rtol=1e-2, atol=1e-10
    )

    # test to check get_residuals()
    cnmfe_residuals = df.iloc[-1].cnmf.get_residuals()
    # something wrong with residuals groundtruth file, maybe it was not created with proper Y - (A * C) - (b * f)
    # cnmfe_residuals_actual = numpy.load(ground_truths_dir.joinpath("cnmfe_full", "cnmfe_residuals.npy"))
    numpy.testing.assert_allclose(
        cnmfe_residuals.as_numpy(),
        df.iloc[-1].caiman.get_input_movie() - cnmfe_reconstructed_movie_AouterC_actual - cnmfe_reconstructed_background_actual,
        rtol=1e2,
        atol=1e-1
    )

    # test to check passing optional ixs components to various functions
    ixs_components = numpy.array([1, 4, 7, 3])

    # test to check ixs components for cnmf.get_masks()
    ixs_spatial_masks = df.iloc[-1].cnmf.get_masks(ixs_components)
    ixs_spatial_masks_actual = numpy.load(
        ground_truths_dir.joinpath("cnmfe_full", "cnmfe_ixs", "ixs_spatial_masks.npy"),
        allow_pickle=True,
    )
    numpy.testing.assert_array_equal(ixs_spatial_masks, ixs_spatial_masks_actual)

    # test to check ixs components for cnmf.get_contours()
    ixs_contours_contours = df.iloc[-1].cnmf.get_contours(ixs_components)[0]
    ixs_contours_contours_actual = numpy.load(
        ground_truths_dir.joinpath(
            "cnmfe_full", "cnmfe_ixs", "ixs_spatial_contours_contours.npy"
        ),
        allow_pickle=True,
    )
    ixs_contours_coms = df.iloc[-1].cnmf.get_contours(ixs_components)[1]
    ixs_contours_coms_actual = numpy.load(
        ground_truths_dir.joinpath(
            "cnmfe_full", "cnmfe_ixs", "ixs_spatial_contours_coms.npy"
        ),
        allow_pickle=True,
    )
    for contour, actual_contour in zip(
            ixs_contours_contours, ixs_contours_contours_actual
    ):
        numpy.testing.assert_allclose(contour, actual_contour, rtol=1e-2, atol=1e-10)
    for com, actual_com in zip(ixs_contours_coms, ixs_contours_coms_actual):
        numpy.testing.assert_allclose(com, actual_com, rtol=1e-2, atol=1e-10)

    # test to check ixs components for cnmf.get_temporal()
    ixs_temporal_components = df.iloc[-1].cnmf.get_temporal(ixs_components)
    ixs_temporal_components_actual = numpy.load(
        ground_truths_dir.joinpath(
            "cnmfe_full", "cnmfe_ixs", "ixs_temporal_components.npy"
        ),
        allow_pickle=True,
    )
    numpy.testing.assert_allclose(
        ixs_temporal_components, ixs_temporal_components_actual, rtol=1e1, atol=1e-5
    )


def test_remove_item():
    set_parent_raw_data_path(vid_dir)
    algo = "mcorr"
    df, batch_path = _create_tmp_batch()
    input_movie_path = get_datafile(algo)

    # make small version of movie for quick testing
    movie = tifffile.imread(input_movie_path)
    small_movie_path = input_movie_path.parent.joinpath("small_movie.tif")
    tifffile.imwrite(small_movie_path, movie[:1001])

    print(input_movie_path)
    df.caiman.add_item(
        algo="mcorr",
        item_name=f"test0",
        input_movie_path=small_movie_path,
        params=test_params["mcorr"],
    )

    df.caiman.add_item(
        algo="mcorr",
        item_name=f"test1",
        input_movie_path=small_movie_path,
        params=test_params["mcorr"],
    )

    df.caiman.add_item(
        algo="cnmf",
        item_name=f"test2",
        input_movie_path=small_movie_path,
        params=test_params["cnmf"],
    )

    diff_params = deepcopy(test_params["cnmf"])
    diff_params["main"]["gSig"] = (6, 6)
    diff_params["main"]["merge_thr"] = 0.85
    df.caiman.add_item(
        algo="cnmf",
        item_name=f"test3",
        input_movie_path=small_movie_path,
        params=diff_params,
    )

    for i, r in df.iterrows():
        proc = r.caiman.run()
        # proc.wait()

    df = load_batch(df.paths.get_batch_path())

    # make sure we can get mcorr movie output of 0th and 1st indices
    path0 = df.iloc[0].mcorr.get_output_path()
    path0_input = df.iloc[0].caiman.get_input_movie_path()
    assert path0.exists()
    df.iloc[0].mcorr.get_output()
    # index 1
    path1 = df.iloc[1].mcorr.get_output_path()
    path1_input = df.iloc[1].caiman.get_input_movie_path()
    assert path1.exists()
    df.iloc[1].mcorr.get_output()

    # make sure we can get cnmf output of 2nd and 3rd indices
    path2 = df.iloc[2].cnmf.get_output_path()
    assert path2.exists()
    df.iloc[2].cnmf.get_output()
    data2 = df.iloc[2].cnmf.get_temporal()
    path3 = df.iloc[3].cnmf.get_output_path()
    assert path3.exists()
    df.iloc[3].cnmf.get_output()
    data3 = df.iloc[3].cnmf.get_temporal()

    # remove index 1
    df.caiman.remove_item(index=1, remove_data=True)
    assert (path1.exists() == False)
    assert df.isin([f"test1"]).any().any() == False
    # input movie path should be unaffected
    assert path1_input.exists()

    # shouldn't affect data at other indices
    assert path0.exists()
    assert df.iloc[0]["item_name"] == f"test0"
    assert df.iloc[1]["item_name"] == f"test2"
    assert df.iloc[2]["item_name"] == f"test3"
    assert df.iloc[0].mcorr.get_output_path().exists()
    df.iloc[0].mcorr.get_output()
    assert df.iloc[1].cnmf.get_output_path().exists()
    df.iloc[1].cnmf.get_output()
    # check that the earlier data from index 2, now index 1, is equal
    np.testing.assert_array_equal(data2, df.iloc[1].cnmf.get_temporal())
    assert df.iloc[2].cnmf.get_output_path().exists()
    df.iloc[2].cnmf.get_output()
    # check that the earlier data from index 3, now index 2, is equal
    np.testing.assert_array_equal(data3,  df.iloc[2].cnmf.get_temporal())
    np.testing.assert_raises(
        AssertionError,
        np.testing.assert_array_equal,
        data2,
        df.iloc[2].cnmf.get_temporal()
    )


def test_cache():
    print("*** Testing cache ***")
    cnmf.cnmf_cache.clear_cache()

    set_parent_raw_data_path(vid_dir)
    algo = "mcorr"

    df, batch_path = _create_tmp_batch()

    batch_path = Path(batch_path)
    batch_dir = batch_path.parent

    input_movie_path = get_datafile(algo)
    print(input_movie_path)

    df.caiman.add_item(
        algo=algo,
        item_name=f"test-{algo}",
        input_movie_path=input_movie_path,
        params=test_params[algo],
    )

    assert df.iloc[-1]["algo"] == algo
    assert df.iloc[-1]["item_name"] == f"test-{algo}"
    assert df.iloc[-1]["params"] == test_params[algo]
    assert df.iloc[-1]["outputs"] is None
    try:
        UUID(df.iloc[-1]["uuid"])
    except:
        pytest.fail("Something wrong with setting UUID for batch items")

    assert vid_dir.joinpath(df.iloc[-1]["input_movie_path"]) == vid_dir.joinpath(
        f"{algo}.tif"
    )

    process = df.iloc[-1].caiman.run()
    # process.wait()

    df = load_batch(batch_path)

    with pd.option_context("display.max_rows", None, "display.max_columns", None):
        print(df)

    pprint(df.iloc[-1]["outputs"], width=-1)
    print(df.iloc[-1]["outputs"]["traceback"])
    assert df.iloc[-1]["outputs"]["success"] is True
    assert df.iloc[-1]["outputs"]["traceback"] is None

    assert (
            batch_dir.joinpath(df.iloc[-1]["outputs"]["mcorr-output-path"])
            == df.paths.resolve(df.iloc[-1]["outputs"]["mcorr-output-path"])
            == batch_dir.joinpath(
        str(df.iloc[-1]["uuid"]),
        f'{df.iloc[-1]["uuid"]}-mcorr_els__d1_60_d2_80_d3_1_order_F_frames_2000.mmap',
    )
    )

    algo = "cnmf"
    input_movie_path = df.iloc[-1].mcorr.get_output_path()
    df.caiman.add_item(
        algo=algo,
        item_name=f"test-{algo}",
        input_movie_path=input_movie_path,
        params=test_params[algo],
    )

    assert df.iloc[-1]["algo"] == algo
    assert df.iloc[-1]["item_name"] == f"test-{algo}"
    assert df.iloc[-1]["params"] == test_params[algo]
    assert df.iloc[-1]["outputs"] is None
    try:
        UUID(df.iloc[-1]["uuid"])
    except:
        pytest.fail("Something wrong with setting UUID for batch items")
    print("cnmf input_movie_path:", df.iloc[-1]["input_movie_path"])
    assert batch_dir.joinpath(df.iloc[-1]["input_movie_path"]) == input_movie_path

    process = df.iloc[-1].caiman.run()
    # process.wait()

    df = load_batch(batch_path)

    with pd.option_context("display.max_rows", None, "display.max_columns", None):
        print(df)

    pprint(df.iloc[-1]["outputs"], width=-1)
    print(df.iloc[-1]["outputs"]["traceback"])

    # test that cache values are returned when calls are made to same function

    # testing that cache size limits work
    cnmf.cnmf_cache.set_maxsize("1M")
    cnmf_output = df.iloc[-1].cnmf.get_output()
    hex_get_output = hex(id(cnmf_output))
    cache = cnmf.cnmf_cache.get_cache()
    hex1 = hex(id(cache[cache["function"] == "get_output"]["return_val"].item()))
    # assert(hex(id(df.iloc[-1].cnmf.get_output(copy=False))) == hex1)
    # assert(hex_get_output != hex1)
    time_stamp1 = cache[cache["function"] == "get_output"]["time_stamp"].item()
    df.iloc[-1].cnmf.get_temporal("good")
    df.iloc[-1].cnmf.get_contours("good")
    df.iloc[-1].cnmf.get_masks("good")
    df.iloc[-1].cnmf.get_temporal(np.arange(7))
    df.iloc[-1].cnmf.get_temporal(np.arange(8))
    df.iloc[-1].cnmf.get_temporal(np.arange(9))
    df.iloc[-1].cnmf.get_temporal(np.arange(6))
    df.iloc[-1].cnmf.get_temporal(np.arange(5))
    df.iloc[-1].cnmf.get_temporal(np.arange(4))
    df.iloc[-1].cnmf.get_temporal(np.arange(3))
    df.iloc[-1].cnmf.get_masks(np.arange(8))
    df.iloc[-1].cnmf.get_masks(np.arange(9))
    df.iloc[-1].cnmf.get_masks(np.arange(7))
    df.iloc[-1].cnmf.get_masks(np.arange(6))
    df.iloc[-1].cnmf.get_masks(np.arange(5))
    df.iloc[-1].cnmf.get_masks(np.arange(4))
    df.iloc[-1].cnmf.get_masks(np.arange(3))
    time_stamp2 = cache[cache["function"] == "get_output"]["time_stamp"].item()
    hex2 = hex(id(cache[cache["function"] == "get_output"]["return_val"].item()))
    assert (cache[cache["function"] == "get_output"].index.size == 1)
    # after adding enough items for cache to exceed max size, cache should remove least recently used items until
    # size is back under max
    assert (len(cnmf.cnmf_cache.get_cache().index) == 17)
    # the time stamp to get_output the second time should be greater than the original time
    # stamp because the cached item is being returned and therefore will have been accessed more recently
    assert (time_stamp2 > time_stamp1)
    # the hex id of the item in the cache when get_output is first called
    # should be the same hex id of the item in the cache when get_output is called again
    assert (hex1 == hex2)

    # test clear_cache()
    cnmf.cnmf_cache.clear_cache()
    assert (len(cnmf.cnmf_cache.get_cache().index) == 0)

    # checking that cache is cleared, checking speed at which item is returned
    start = time.time()
    df.iloc[-1].cnmf.get_output()
    end = time.time()
    assert (len(cnmf.cnmf_cache.get_cache().index) == 1)

    # second call to item now added to cache, time to return item should be must faster than before because item has
    # now been cached
    start2 = time.time()
    df.iloc[-1].cnmf.get_output()
    end2 = time.time()
    assert(end2-start2 < end-start)

    # testing clear_cache() again, length of dataframe should be zero
    cnmf.cnmf_cache.clear_cache()
    assert (len(cnmf.cnmf_cache.get_cache().index) == 0)

    # test setting maxsize as 0, should effectively disable the cache...additionally, time to return an item called
    # twice should roughly be the same because item is not being stored in the cache
    # cache length should remain zero throughout calls to extension functions
    cnmf.cnmf_cache.set_maxsize(0)
    start = time.time()
    df.iloc[-1].cnmf.get_output()
    end = time.time()
    assert (len(cnmf.cnmf_cache.get_cache().index) == 0)

    start2 = time.time()
    df.iloc[-1].cnmf.get_output()
    end2 = time.time()
    assert (len(cnmf.cnmf_cache.get_cache().index) == 0)
    assert(abs((end-start)-(end2-start2)) < 0.05)

    # test to check that separate cache items are being returned for different batch items
    # must add another item to the batch, running cnmfe

    input_movie_path = get_datafile("cnmfe")
    print(input_movie_path)
    df.caiman.add_item(
        algo="mcorr",
        item_name=f"test-cnmfe-mcorr",
        input_movie_path=input_movie_path,
        params=test_params["mcorr"],
    )
    process = df.iloc[-1].caiman.run()
    # process.wait()

    df = load_batch(batch_path)

    algo = "cnmfe"
    param_name = "cnmfe_full"
    input_movie_path = df.iloc[-1].mcorr.get_output_path()
    print(input_movie_path)

    df.caiman.add_item(
        algo=algo,
        item_name=f"test-{algo}",
        input_movie_path=input_movie_path,
        params=test_params[param_name],
    )

    process = df.iloc[-1].caiman.run()
    # process.wait()

    df = load_batch(batch_path)

    cnmf.cnmf_cache.set_maxsize("1M")

    df.iloc[1].cnmf.get_output() # cnmf output
    df.iloc[-1].cnmf.get_output() # cnmfe output

    cache = cnmf.cnmf_cache.get_cache()

    # checking that both outputs from different batch items are added to the cache
    assert(len(cache.index) == 2)

    # checking that the uuid of each outputs from the different batch items are not the same
    assert(cache.iloc[-1]["uuid"] != cache.iloc[-2]["uuid"])

    # checking that the uuid of the output in the cache is the correct uuid of the batch item in the df
    assert(cache.iloc[-1]["uuid"] == df.iloc[-1]["uuid"])

    # call get output from cnmf, check that it is the most recent thing called in the cache
    df.iloc[1].cnmf.get_output()
    cnmf_uuid = df.iloc[1]["uuid"]
    most_recently_called = cache.sort_values(by=["time_stamp"], ascending=True).iloc[-1]
    cache_uuid = most_recently_called["uuid"]
    assert(cnmf_uuid == cache_uuid)

    # check to make sure by certain params that it is cnmf vs cnmfe
    output = df.iloc[1].cnmf.get_output()
    assert(output.params.patch["low_rank_background"] == True)
    output2 = df.iloc[-1].cnmf.get_output()
    assert(output2.params.patch["low_rank_background"] == False)
    
    # test for copy
    # if return_copy=True, then hex id of calls to the same function should be false
    output = df.iloc[1].cnmf.get_output()
    assert(hex(id(output)) != hex(id(cache.sort_values(by=["time_stamp"], ascending=True).iloc[-1])))
    # if return_copy=False, then hex id of calls to the same function should be true
    output = df.iloc[1].cnmf.get_output(return_copy=False)
    output2 = df.iloc[1].cnmf.get_output(return_copy=False)
    assert(hex(id(output)) == hex(id(output2)))
    assert(hex(id(cnmf.cnmf_cache.get_cache().iloc[-1]["return_val"])) == hex(id(output)))



"""Performs CNMF in a separate process"""
import click
import caiman as cm
from caiman.source_extraction.cnmf import cnmf as cnmf
from caiman.source_extraction.cnmf.params import CNMFParams
import psutil
import numpy as np
import pandas as pd
import traceback
from pathlib import Path
from shutil import move as move_file
import os
import time
from datetime import datetime

# prevent circular import
if __name__ in ["__main__", "__mp_main__"]:  # when running in subprocess
    from mesmerize_core import set_parent_raw_data_path, load_batch
    from mesmerize_core.utils import IS_WINDOWS
else:  # when running with local backend
    from ..batch_utils import set_parent_raw_data_path, load_batch
    from ..utils import IS_WINDOWS


def run_algo(batch_path, uuid, data_path: str = None):
    algo_start = time.time()
    set_parent_raw_data_path(data_path)

    df = load_batch(batch_path)
    item = df[df["uuid"] == uuid].squeeze()

    input_movie_path = item["input_movie_path"]
    # resolve full path
    input_movie_path = str(df.paths.resolve(input_movie_path))

    # make output dir
    output_dir = Path(batch_path).parent.joinpath(str(uuid))
    output_dir.mkdir(parents=True, exist_ok=True)

    params = item["params"]
    print(
        f"************************************************************************\n\n"
        f"Starting CNMF item:\n{item}\nWith params:{params}"
    )

    # adapted from current demo notebook
    if "MESMERIZE_N_PROCESSES" in os.environ.keys():
        try:
            n_processes = int(os.environ["MESMERIZE_N_PROCESSES"])
        except:
            n_processes = psutil.cpu_count() - 1
    else:
        n_processes = psutil.cpu_count() - 1
    # Start cluster for parallel processing
    c, dview, n_processes = cm.cluster.setup_cluster(
        backend="local", n_processes=n_processes, single_thread=False
    )

    # merge cnmf and eval kwargs into one dict
    cnmf_params = CNMFParams(params_dict=params["main"])
    # Run CNMF, denote boolean 'success' if CNMF completes w/out error
    try:
        fname_new = cm.save_memmap(
            [input_movie_path], base_name=f"{uuid}_cnmf-memmap_", order="C", dview=dview
        )

        print("making memmap")

        Yr, dims, T = cm.load_memmap(fname_new)
        images = np.reshape(Yr.T, [T] + list(dims), order="F")

        proj_paths = dict()
        for proj_type in ["mean", "std", "max"]:
            p_img = getattr(np, f"nan{proj_type}")(images, axis=0)
            proj_paths[proj_type] = output_dir.joinpath(
                f"{uuid}_{proj_type}_projection.npy"
            )
            np.save(str(proj_paths[proj_type]), p_img)

        # in fname new load in memmap order C
        cm.stop_server(dview=dview)
        c, dview, n_processes = cm.cluster.setup_cluster(
            backend="local", n_processes=None, single_thread=False
        )

        print("performing CNMF")
        cnm = cnmf.CNMF(n_processes, params=cnmf_params, dview=dview)

        print("fitting images")
        cnm = cnm.fit(images)
        #
        if "refit" in params.keys():
            if params["refit"] is True:
                print("refitting")
                cnm = cnm.refit(images, dview=dview)

        print("performing eval")
        cnm.estimates.evaluate_components(images, cnm.params, dview=dview)

        output_path = output_dir.joinpath(f"{uuid}.hdf5").resolve()

        cnm.save(str(output_path))

        Cn = cm.local_correlations(images.transpose(1, 2, 0))
        Cn[np.isnan(Cn)] = 0

        corr_img_path = output_dir.joinpath(f"{uuid}_cn.npy").resolve()
        np.save(str(corr_img_path), Cn, allow_pickle=False)

        # output dict for dataframe row (pd.Series)
        d = dict()

        cnmf_memmap_path = output_dir.joinpath(Path(fname_new).name)
        if IS_WINDOWS:
            Yr._mmap.close()  # accessing private attr but windows is annoying otherwise
        move_file(fname_new, cnmf_memmap_path)

        cnmf_hdf5_path = output_path.relative_to(output_dir.parent)
        cnmf_memmap_path = cnmf_memmap_path.relative_to(output_dir.parent)
        corr_img_path = corr_img_path.relative_to(output_dir.parent)
        for proj_type in proj_paths.keys():
            d[f"{proj_type}-projection-path"] = proj_paths[proj_type].relative_to(
                output_dir.parent
            )

        d.update(
            {
                "cnmf-hdf5-path": cnmf_hdf5_path,
                "cnmf-memmap-path": cnmf_memmap_path,
                "corr-img-path": corr_img_path,
                "success": True,
                "traceback": None,
            }
        )

    except:
        d = {"success": False, "traceback": traceback.format_exc()}

    cm.stop_server(dview=dview)

    # Add dictionary to output column of series
    df.loc[df["uuid"] == uuid, "outputs"] = [d]
    # Add ran timestamp to ran_time column of series
    df.loc[df["uuid"] == uuid, "ran_time"] = datetime.now().isoformat(timespec="seconds", sep="T")
    df.loc[df["uuid"] == uuid, "algo_duration"] = str(round(time.time() - algo_start, 2)) + " sec"
    # save dataframe to disc
    df.to_pickle(batch_path)


@click.command()
@click.option("--batch-path", type=str)
@click.option("--uuid", type=str)
@click.option("--data-path")
def main(batch_path, uuid, data_path: str = None):
    run_algo(batch_path, uuid, data_path)


if __name__ == "__main__":
    main()



from functools import wraps
from typing import Union
from uuid import UUID

from mesmerize_core.caiman_extensions._batch_exceptions import BatchItemNotRunError, BatchItemUnsuccessfulError, \
    WrongAlgorithmExtensionError


def validate(algo: str = None):
    def dec(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if self._series["outputs"] is None:
                raise BatchItemNotRunError("Item has not been run")

            if algo is not None:
                if algo not in self._series["algo"]:
                    raise WrongAlgorithmExtensionError(
                        f"<{algo}> extension called for a <{self._series.algo}> item"
                    )

            if not self._series["outputs"]["success"]:
                tb = self._series["outputs"]["traceback"]
                raise BatchItemUnsuccessfulError(f"Batch item was unsuccessful, traceback from subprocess:\n{tb}")
            return func(self, *args, **kwargs)

        return wrapper

    return dec


def _index_parser(func):
    @wraps(func)
    def _parser(instance, *args, **kwargs):
        if "index" in kwargs.keys():
            index: Union[int, str, UUID] = kwargs["index"]
        elif len(args) > 0:
            index = args[0]  # always first positional arg

        if isinstance(index, (UUID, str)):
            _index = instance._df[instance._df["uuid"] == str(index)].index
            if _index.size == 0:
                raise ValueError(f"No batch item found with uuid: {index}")

            index = _index.item()

        if not isinstance(index, int):
            raise TypeError(f"`index` argument must be of type `int`, `str`, or `UUID`")

        if "index" in kwargs.keys():
            kwargs["index"] = index
        else:
            args = (index, *args[1:])

        return func(instance, *args, **kwargs)
    return _parser



"""
Useful functions adapted from old mesmerize

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""


import numpy as np
from functools import wraps
import os
from stat import S_IEXEC
from typing import *
import re as regex
from pathlib import Path
from warnings import warn
import sys
from tempfile import NamedTemporaryFile
from subprocess import check_call

if os.name == "nt":
    IS_WINDOWS = True
    HOME = "USERPROFILE"
else:
    IS_WINDOWS = False
    HOME = "HOME"

if "MESMERIZE_LRU_CACHE" in os.environ.keys():
    MESMERIZE_LRU_CACHE = os.environ["MESMERIZE_LRU_CACHE"]
else:
    MESMERIZE_LRU_CACHE = 10


def warning_experimental(more_info: str = ""):
    """
    decorator to warn the user that the function is experimental
    """
    def catcher(func):
        @wraps(func)
        def fn(self, *args, **kwargs):
            warn(
                f"You are trying to use the following experimental feature, "
                f"this may change in the future without warning:\n"
                f"{func.__qualname__}\n"
                f"{more_info}\n",
                FutureWarning,
                stacklevel=2
            )
            return func(self, *args, **kwargs)
        return fn
    return catcher


def validate_path(path: Union[str, Path]):
    if not regex.match("^[A-Za-z0-9@\/\\\:._-]*$", str(path)):
        raise ValueError(
            "Paths must only contain alphanumeric characters, "
            "hyphens ( - ), underscores ( _ ) or periods ( . )"
        )
    return path


def make_runfile(
    module_path: str, args_str: Optional[str] = None, filename: Optional[str] = None
) -> str:
    """
    Make an executable bash script.
    Used for running python scripts in external processes within the same python environment as the main/parent process.

    Parameters
    ----------
    module_path: str
        absolute path to the python module/script that should be run externally

    args_str: Optional[str]
        optinal str of args that is directly passed to the script specified by ``module_path``

    filename: Optional[str]
        optional, filename of the executable bash script

    Returns
    -------
    str
        path to the shell script that can be executed
    """

    if filename is None:
        if IS_WINDOWS:
            sh_file = os.path.join(os.environ[HOME], "run.ps1")
        else:
            sh_file = os.path.join(os.environ[HOME], "run.sh")
    else:
        if IS_WINDOWS:
            if not filename.endswith(".ps1"):
                filename = filename + ".ps1"

    sh_file = filename

    if args_str is None:
        args_str = ""

    if not IS_WINDOWS:
        with open(sh_file, "w") as f:

            f.write(f"#!/bin/bash\n")

            if "CONDA_PREFIX" in os.environ.keys():
                f.write(
                    f'export CONDA_PREFIX={os.environ["CONDA_PREFIX"]}\n'
                    f'export CONDA_PYTHON_EXE={os.environ["CONDA_PYTHON_EXE"]}\n'
                    f'export CONDA_PREFIX_1={os.environ["CONDA_PREFIX_1"]}\n'
                )

            elif "VIRTUAL_ENV" in os.environ.keys():
                f.write(
                    f'export PATH={os.environ["PATH"]}\n'
                    f'export VIRTUAL_ENV={os.environ["VIRTUAL_ENV"]}\n'
                    f'export LD_LIBRARY_PATH={os.environ["LD_LIBRARY_PATH"]}\n'
                )

            if "PYTHONPATH" in os.environ.keys():
                f.write(f'export PYTHONPATH={os.environ["PYTHONPATH"]}\n')

            # for k, v in os.environ.items():  # copy the current environment
            #     if '\n' in v:
            #         continue
            #
            # f.write(f'export {k}="{v}"\n')

            # User-setable n-processes
            if "MESMERIZE_N_PROCESSES" in os.environ.keys():
                f.write(
                    f'export MESMERIZE_N_PROCESSES={os.environ["MESMERIZE_N_PROCESSES"]}\n'
                )

            f.write(f"export OPENBLAS_NUM_THREADS=1\n" f"export MKL_NUM_THREADS=1\n")

            f.write(f"python {module_path} {args_str}")  # call the script to run

    else:
        with open(sh_file, "w") as f:
            for k, v in os.environ.items():  # copy the current environment
                if regex.match("^.*[\(\)]", str(k)) or regex.match("^.*[\(\)]", str(v)):
                    continue
                with NamedTemporaryFile(suffix=".ps1", delete=False) as tmp:
                    try:  # windows powershell is stupid so make sure all the env var names work
                        tmp.write(f'$env:{k}="{v}";\n')
                        tmp.close()
                        check_call(f"powershell {tmp.name}")
                        os.unlink(tmp.name)
                    except:
                        continue
                f.write(f'$env:{k}="{v}";\n')  # write only env vars that powershell likes
            f.write(f"{sys.executable} {module_path} {args_str}")

    st = os.stat(sh_file)
    os.chmod(sh_file, st.st_mode | S_IEXEC)

    return sh_file


def quick_min_max(data: np.ndarray) -> Tuple[float, float]:
    # from pyqtgraph.ImageView
    # Estimate the min/max values of *data* by subsampling.
    # Returns [(min, max), ...] with one item per channel
    while data.size > 1e6:
        ax = np.argmax(data.shape)
        sl = [slice(None)] * data.ndim
        sl[ax] = slice(None, None, 2)
        data = data[tuple(sl)]

    return float(np.nanmin(data)), float(np.nanmax(data))


def _organize_coordinates(contour: dict):
    coors = contour["coordinates"]
    coors = coors[~np.isnan(coors).any(axis=1)]

    return coors
